"""
Builds bootable CD images that have PXE-equivalent behavior
for all Cobbler distros/profiles/systems currently in memory.

Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import logging
import os
import os.path
import re
import shutil
from typing import Optional

from cobbler import utils


class BuildIso:
    """
    Handles conversion of internal state to the isolinux tree layout
    """
    def __init__(self, collection_mgr, verbose: bool = False):
        """
        Constructor
        """
        self.verbose = verbose
        self.collection_mgr = collection_mgr
        self.settings = collection_mgr.settings()
        self.api = collection_mgr.api
        self.distros = collection_mgr.distros()
        self.profiles = collection_mgr.profiles()
        self.systems = collection_mgr.systems()
        self.distros = collection_mgr.distros()
        self.distmap = {}
        self.distctr = 0
        self.source = ""
        self.logger = logging.getLogger()
        # grab the header from buildiso.header file
        header_src = open(os.path.join(self.settings.iso_template_dir, "buildiso.template"))
        self.iso_template = header_src.read()
        header_src.close()

    def add_remaining_kopts(self, koptdict: dict) -> str:
        """
        Add remaining kernel_options to append_line

        :param koptdict: The kernel options which are not present in append_line.
        :return: A single line with all kernel options
        """
        append_line = ""
        for (k, v) in list(koptdict.items()):
            if v is None:
                append_line += " %s" % k
            else:
                if isinstance(v, list):
                    for i in v:
                        _i = str(i).strip()
                        if ' ' in _i:
                            append_line += " %s='%s'" % (k, _i)
                        else:
                            append_line += " %s=%s" % (k, _i)
                else:
                    _v = str(v).strip()
                    if ' ' in _v:
                        append_line += " %s='%s'" % (k, _v)
                    else:
                        append_line += " %s=%s" % (k, _v)
        append_line += "\n"
        return append_line

    def make_shorter(self, distname: str) -> str:
        """
        Return A short distro identifier.

        :param distname: The distro name to return a identifier for.
        :return: A short distro identifier
        """
        if distname in self.distmap:
            return self.distmap[distname]
        else:
            self.distctr += 1
            self.distmap[distname] = str(self.distctr)
            return str(self.distctr)

    def copy_boot_files(self, distro, destdir, prefix: Optional[str] = None):
        """
        Copy kernel/initrd to destdir with (optional) newfile prefix

        :param distro: Distro object to return the boot files for.
        :param destdir: The destionation direcotry.
        :param prefix: The file prefix.
        """
        if not os.path.exists(distro.kernel):
            utils.die("path does not exist: %s" % distro.kernel)
        if not os.path.exists(distro.initrd):
            utils.die("path does not exist: %s" % distro.initrd)
        if prefix is None:
            shutil.copyfile(distro.kernel, os.path.join(destdir, "%s" % os.path.basename(distro.kernel)))
            shutil.copyfile(distro.initrd, os.path.join(destdir, "%s" % os.path.basename(distro.initrd)))
        else:
            shutil.copyfile(distro.kernel, os.path.join(destdir, "%s.krn" % prefix))
            shutil.copyfile(distro.initrd, os.path.join(destdir, "%s.img" % prefix))

    def filter_systems_or_profiles(self, selected_items, list_type: str) -> list:
        """
        Return a list of valid profile or system objects selected from all profiles or systems by name, or everything if
        selected_items is empty.

        :param selected_items: The filter to match certain objects with. The filter will be applied to the object name.
        :param list_type: Must be "profile" or "system".
        :return: A list of valid profiles OR systems.
        """
        if list_type == 'profile':
            all_objs = [profile for profile in self.api.profiles()]
        elif list_type == 'system':
            all_objs = [system for system in self.api.systems()]
        else:
            utils.die("Invalid list_type argument: " + list_type)

        all_objs.sort(key=lambda profile: profile.name)

        # no profiles/systems selection is made, let's process everything
        if not selected_items:
            return all_objs

        which_objs = []
        selected_list = utils.input_string_or_list(selected_items)
        for obj in all_objs:
            if obj.name in selected_list:
                which_objs.append(obj)
                selected_list.remove(obj.name)

        for bad_name in selected_list:
            self.logger.warning("WARNING: %s is not a valid %s" % (bad_name, list_type))

        if not which_objs:
            utils.die("No valid systems or profiles were specified.")

        return which_objs

    def generate_netboot_iso(self, imagesdir, isolinuxdir, profiles=None, systems=None,
                             exclude_dns: Optional[bool] = None):
        """
        Create bootable CD image to be used for network installations

        :param imagesdir: Currently unused parameter.
        :param isolinuxdir: The parent directory where the isolinux.cfg is located.
        :param profiles: The filter to generate a netboot iso for. You may specify multiple profiles on the CLI space
                        separated.
        :param systems: The filter to generate a netboot iso for. You may specify multiple systems on the CLI space
                        separated.
        :param exclude_dns: If this is True then the dns server is skipped. None or False will set it.
        """
        which_profiles = self.filter_systems_or_profiles(profiles, 'profile')
        which_systems = self.filter_systems_or_profiles(systems, 'system')

        # setup isolinux.cfg
        isolinuxcfg = os.path.join(isolinuxdir, "isolinux.cfg")
        cfg = open(isolinuxcfg, "w+")
        cfg.write(self.iso_template)

        # iterate through selected profiles
        for profile in which_profiles:
            self.logger.info("processing profile: %s" % profile.name)
            dist = profile.get_conceptual_parent()
            distname = self.make_shorter(dist.name)
            self.copy_boot_files(dist, isolinuxdir, distname)

            cfg.write("\n")
            cfg.write("LABEL %s\n" % profile.name)
            cfg.write("  MENU LABEL %s\n" % profile.name)
            cfg.write("  kernel %s.krn\n" % distname)

            data = utils.blender(self.api, False, profile)

            # SUSE is not using 'text'. Instead 'textmode' is used as kernel option.
            if dist is not None:
                utils.kopts_overwrite(data['kernel_options'], self.api.settings().server, dist.breed)

            if not re.match(r"[a-z]+://.*", data["autoinstall"]):
                data["autoinstall"] = "http://%s:%s/cblr/svc/op/autoinstall/profile/%s" % (
                    data["server"], self.api.settings().http_port, profile.name
                )

            append_line = " append initrd=%s.img" % distname
            if dist.breed == "suse":
                if "proxy" in data and data["proxy"] != "":
                    append_line += " proxy=%s" % data["proxy"]
                if "install" in data["kernel_options"] and data["kernel_options"]["install"] != "":
                    v = data["kernel_options"]["install"]
                    if isinstance(v, list):
                        v = v[0]
                        append_line += " install=%s" % v
                    del data["kernel_options"]["install"]
                else:
                    append_line += " install=http://%s:%s/cblr/links/%s" % (
                        data["server"], self.api.settings().http_port, dist.name
                    )
                if "autoyast" in data["kernel_options"] and data["kernel_options"]["autoyast"] != "":
                    append_line += " autoyast=%s" % data["kernel_options"]["autoyast"]
                    del data["kernel_options"]["autoyast"]
                else:
                    append_line += " autoyast=%s" % data["autoinstall"]

            if dist.breed == "redhat":
                if "proxy" in data and data["proxy"] != "":
                    append_line += " proxy=%s http_proxy=%s" % (data["proxy"], data["proxy"])
                append_line += " inst.ks=%s" % data["autoinstall"]

            if dist.breed in ["ubuntu", "debian"]:
                append_line += " auto-install/enable=true url=%s" % data["autoinstall"]
                if "proxy" in data and data["proxy"] != "":
                    append_line += " mirror/http/proxy=%s" % data["proxy"]
            append_line += self.add_remaining_kopts(data["kernel_options"])
            cfg.write(append_line)

        cfg.write("\nMENU SEPARATOR\n")

        # iterate through all selected systems
        for system in which_systems:
            self.logger.info("processing system: %s" % system.name)
            profile = system.get_conceptual_parent()
            dist = profile.get_conceptual_parent()
            distname = self.make_shorter(dist.name)
            self.copy_boot_files(dist, isolinuxdir, distname)

            cfg.write("\n")
            cfg.write("LABEL %s\n" % system.name)
            cfg.write("  MENU LABEL %s\n" % system.name)
            cfg.write("  kernel %s.krn\n" % distname)

            data = utils.blender(self.api, False, system)
            if not re.match(r"[a-z]+://.*", data["autoinstall"]):
                data["autoinstall"] = "http://%s:%s/cblr/svc/op/autoinstall/system/%s" % (
                    data["server"], self.api.settings().http_port, system.name
                )

            append_line = " append initrd=%s.img" % distname
            if dist.breed == "suse":
                if "proxy" in data and data["proxy"] != "":
                    append_line += " proxy=%s" % data["proxy"]
                if "install" in data["kernel_options"] and data["kernel_options"]["install"] != "":
                    append_line += " install=%s" % data["kernel_options"]["install"]
                    del data["kernel_options"]["install"]
                else:
                    append_line += " install=http://%s:%s/cblr/links/%s" % (
                        data["server"], self.api.settings().http_port, dist.name
                    )
                if "autoyast" in data["kernel_options"] and data["kernel_options"]["autoyast"] != "":
                    append_line += " autoyast=%s" % data["kernel_options"]["autoyast"]
                    del data["kernel_options"]["autoyast"]
                else:
                    append_line += " autoyast=%s" % data["autoinstall"]

            if dist.breed == "redhat":
                if "proxy" in data and data["proxy"] != "":
                    append_line += " proxy=%s http_proxy=%s" % (data["proxy"], data["proxy"])
                append_line += " inst.ks=%s" % data["autoinstall"]

            if dist.breed in ["ubuntu", "debian"]:
                append_line += " auto-install/enable=true url=%s netcfg/disable_autoconfig=true" % data["autoinstall"]
                if "proxy" in data and data["proxy"] != "":
                    append_line += " mirror/http/proxy=%s" % data["proxy"]
                # hostname is required as a parameter, the one in the preseed is not respected
                my_domain = "local.lan"
                if system.hostname != "":
                    # if this is a FQDN, grab the first bit
                    my_hostname = system.hostname.split(".")[0]
                    _domain = system.hostname.split(".")[1:]
                    if _domain:
                        my_domain = ".".join(_domain)
                else:
                    my_hostname = system.name.split(".")[0]
                    _domain = system.name.split(".")[1:]
                    if _domain:
                        my_domain = ".".join(_domain)
                # At least for debian deployments configured for DHCP networking this values are not used, but
                # specifying here avoids questions
                append_line += " hostname=%s domain=%s" % (my_hostname, my_domain)
                # A similar issue exists with suite name, as installer requires the existence of "stable" in the dists
                # directory
                append_line += " suite=%s" % dist.os_version

            # Try to add static ip boot options to avoid DHCP (interface/ip/netmask/gw/dns)
            # Check for overrides first and clear them from kernel_options
            my_int = None
            my_ip = None
            my_mask = None
            my_gw = None
            my_dns = None
            if dist.breed in ["suse", "redhat"]:
                if "netmask" in data["kernel_options"] and data["kernel_options"]["netmask"] != "":
                    my_mask = data["kernel_options"]["netmask"]
                    del data["kernel_options"]["netmask"]
                if "gateway" in data["kernel_options"] and data["kernel_options"]["gateway"] != "":
                    my_gw = data["kernel_options"]["gateway"]
                    del data["kernel_options"]["gateway"]

            if dist.breed == "redhat":
                if "ksdevice" in data["kernel_options"] and data["kernel_options"]["ksdevice"] != "":
                    my_int = data["kernel_options"]["ksdevice"]
                    if my_int == "bootif":
                        my_int = None
                    del data["kernel_options"]["ksdevice"]
                if "ip" in data["kernel_options"] and data["kernel_options"]["ip"] != "":
                    my_ip = data["kernel_options"]["ip"]
                    del data["kernel_options"]["ip"]
                if "dns" in data["kernel_options"] and data["kernel_options"]["dns"] != "":
                    my_dns = data["kernel_options"]["dns"]
                    del data["kernel_options"]["dns"]

            if dist.breed == "suse":
                if "netdevice" in data["kernel_options"] and data["kernel_options"]["netdevice"] != "":
                    my_int = data["kernel_options"]["netdevice"]
                    del data["kernel_options"]["netdevice"]
                if "hostip" in data["kernel_options"] and data["kernel_options"]["hostip"] != "":
                    my_ip = data["kernel_options"]["hostip"]
                    del data["kernel_options"]["hostip"]
                if "nameserver" in data["kernel_options"] and data["kernel_options"]["nameserver"] != "":
                    my_dns = data["kernel_options"]["nameserver"]
                    del data["kernel_options"]["nameserver"]

            if dist.breed in ["ubuntu", "debian"]:
                if "netcfg/choose_interface" in data["kernel_options"] and data["kernel_options"]["netcfg/choose_interface"] != "":
                    my_int = data["kernel_options"]["netcfg/choose_interface"]
                    del data["kernel_options"]["netcfg/choose_interface"]
                if "netcfg/get_ipaddress" in data["kernel_options"] and data["kernel_options"]["netcfg/get_ipaddress"] != "":
                    my_ip = data["kernel_options"]["netcfg/get_ipaddress"]
                    del data["kernel_options"]["netcfg/get_ipaddress"]
                if "netcfg/get_netmask" in data["kernel_options"] and data["kernel_options"]["netcfg/get_netmask"] != "":
                    my_mask = data["kernel_options"]["netcfg/get_netmask"]
                    del data["kernel_options"]["netcfg/get_netmask"]
                if "netcfg/get_gateway" in data["kernel_options"] and data["kernel_options"]["netcfg/get_gateway"] != "":
                    my_gw = data["kernel_options"]["netcfg/get_gateway"]
                    del data["kernel_options"]["netcfg/get_gateway"]
                if "netcfg/get_nameservers" in data["kernel_options"] and data["kernel_options"]["netcfg/get_nameservers"] != "":
                    my_dns = data["kernel_options"]["netcfg/get_nameservers"]
                    del data["kernel_options"]["netcfg/get_nameservers"]

            # If no kernel_options overrides are present find the management interface do nothing when zero or multiple
            # management interfaces are found
            if my_int is None:
                mgmt_ints = []
                mgmt_ints_multi = []
                slave_ints = []
                if len(list(data["interfaces"].keys())) >= 1:
                    for (iname, idata) in list(data["interfaces"].items()):
                        if idata["management"] and idata["interface_type"] in ["bond", "bridge"]:
                            # bonded/bridged management interface
                            mgmt_ints_multi.append(iname)
                        if idata["management"] and idata["interface_type"] not in ["bond", "bridge", "bond_slave", "bridge_slave", "bonded_bridge_slave"]:
                            # single management interface
                            mgmt_ints.append(iname)

                if len(mgmt_ints_multi) == 1 and len(mgmt_ints) == 0:
                    # Bonded/bridged management interface, find a slave interface if eth0 is a slave use that (it's what
                    # people expect)
                    for (iname, idata) in list(data["interfaces"].items()):
                        if idata["interface_type"] in ["bond_slave", "bridge_slave", "bonded_bridge_slave"] and idata["interface_master"] == mgmt_ints_multi[0]:
                            slave_ints.append(iname)

                    if "eth0" in slave_ints:
                        my_int = "eth0"
                    else:
                        my_int = slave_ints[0]
                    # Set my_ip from the bonded/bridged interface here
                    my_ip = data["ip_address_" + data["interface_master_" + my_int]]
                    my_mask = data["netmask_" + data["interface_master_" + my_int]]

                if len(mgmt_ints) == 1 and len(mgmt_ints_multi) == 0:
                    # Single management interface
                    my_int = mgmt_ints[0]

            # Lookup tcp/ip configuration data
            if my_ip is None and my_int is not None:
                intip = "ip_address_" + my_int
                if intip in data and data[intip] != "":
                    my_ip = data["ip_address_" + my_int]

            if my_mask is None and my_int is not None:
                intmask = "netmask_" + my_int
                if intmask in data and data[intmask] != "":
                    my_mask = data["netmask_" + my_int]

            if my_gw is None:
                if "gateway" in data and data["gateway"] != "":
                    my_gw = data["gateway"]

            if my_dns is None:
                if "name_servers" in data and data["name_servers"] != "":
                    my_dns = data["name_servers"]

            # Add information to the append_line
            if my_int is not None:
                intmac = "mac_address_" + my_int
                if dist.breed == "suse":
                    if intmac in data and data[intmac] != "":
                        append_line += " netdevice=%s" % data["mac_address_" + my_int].lower()
                    else:
                        append_line += " netdevice=%s" % my_int
                if dist.breed == "redhat":
                    if intmac in data and data[intmac] != "":
                        append_line += " ksdevice=%s" % data["mac_address_" + my_int]
                    else:
                        append_line += " ksdevice=%s" % my_int
                if dist.breed in ["ubuntu", "debian"]:
                    append_line += " netcfg/choose_interface=%s" % my_int

            if my_ip is not None:
                if dist.breed == "suse":
                    append_line += " hostip=%s" % my_ip
                if dist.breed == "redhat":
                    append_line += " ip=%s" % my_ip
                if dist.breed in ["ubuntu", "debian"]:
                    append_line += " netcfg/get_ipaddress=%s" % my_ip

            if my_mask is not None:
                if dist.breed in ["suse", "redhat"]:
                    append_line += " netmask=%s" % my_mask
                if dist.breed in ["ubuntu", "debian"]:
                    append_line += " netcfg/get_netmask=%s" % my_mask

            if my_gw is not None:
                if dist.breed in ["suse", "redhat"]:
                    append_line += " gateway=%s" % my_gw
                if dist.breed in ["ubuntu", "debian"]:
                    append_line += " netcfg/get_gateway=%s" % my_gw

            if exclude_dns is None or my_dns is not None:
                if dist.breed == "suse":
                    if type(my_dns) == list:
                        append_line += " nameserver=%s" % ",".join(my_dns)
                    else:
                        append_line += " nameserver=%s" % my_dns
                if dist.breed == "redhat":
                    if type(my_dns) == list:
                        append_line += " dns=%s" % ",".join(my_dns)
                    else:
                        append_line += " dns=%s" % my_dns
                if dist.breed in ["ubuntu", "debian"]:
                    if type(my_dns) == list:
                        append_line += " netcfg/get_nameservers=%s" % ",".join(my_dns)
                    else:
                        append_line += " netcfg/get_nameservers=%s" % my_dns

            # Add remaining kernel_options to append_line
            append_line += self.add_remaining_kopts(data["kernel_options"])
            cfg.write(append_line)

        cfg.write("\n")
        cfg.write("MENU END\n")
        cfg.close()

    def generate_standalone_iso(self, imagesdir, isolinuxdir, distname, filesource, airgapped: bool, profiles):
        """
        Create bootable CD image to be used for handsoff CD installtions

        :param imagesdir: Unused Parameter.
        :param isolinuxdir: The parent directory where the file isolinux.cfg is located at.
        :param distname: The name of the Cobbler distribution.
        :param filesource: Not clear what this exactly does
        :param airgapped: Whether the repositories have to be locally available or the internet is reachable.
        :param profiles: The list of profiles to include.
        """
        # Get the distro object for the requested distro and then get all of its descendants (profiles/sub-profiles/
        # systems) with sort=True for profile/system heirarchy to allow menu indenting
        distro = self.api.find_distro(distname)
        if distro is None:
            utils.die("distro %s was not found, aborting" % distname)
        descendants = distro.get_descendants(sort=True)
        profiles = utils.input_string_or_list(profiles)

        if filesource is None:
            # Try to determine the source from the distro kernel path
            self.logger.debug("trying to locate source for distro")
            found_source = False
            (source_head, source_tail) = os.path.split(distro.kernel)
            while source_tail != '':
                if source_head == os.path.join(self.api.settings().webdir, "distro_mirror"):
                    filesource = os.path.join(source_head, source_tail)
                    found_source = True
                    self.logger.debug("found source in %s" % filesource)
                    break
                (source_head, source_tail) = os.path.split(source_head)
            # Can't find the source, raise an error
            if not found_source:
                utils.die("Error, no installation source found. When building a standalone ISO, you must specify a "
                          "--source if the distro install tree is not hosted locally")

        self.logger.info("copying kernels and initrds for standalone distro")
        self.copy_boot_files(distro, isolinuxdir, None)

        self.logger.info("generating an isolinux.cfg")
        isolinuxcfg = os.path.join(isolinuxdir, "isolinux.cfg")
        cfg = open(isolinuxcfg, "w+")
        cfg.write(self.iso_template)

        if airgapped:
            repo_names_to_copy = {}

        for descendant in descendants:
            # if a list of profiles was given, skip any others and their systems
            if profiles and ((descendant.COLLECTION_TYPE == 'profile' and descendant.name not in profiles)
                             or (descendant.COLLECTION_TYPE == 'system' and descendant.profile not in profiles)):
                continue

            menu_indent = 0
            if descendant.COLLECTION_TYPE == 'system':
                menu_indent = 4

            data = utils.blender(self.api, False, descendant)

            # SUSE is not using 'text'. Instead 'textmode' is used as kernel option.
            if distro is not None:
                utils.kopts_overwrite(data['kernel_options'], self.settings.server, distro.breed)

            cfg.write("\n")
            cfg.write("LABEL %s\n" % descendant.name)
            if menu_indent:
                cfg.write("  MENU INDENT %d\n" % menu_indent)
            cfg.write("  MENU LABEL %s\n" % descendant.name)
            cfg.write("  kernel %s\n" % os.path.basename(distro.kernel))

            append_line = "  append initrd=%s" % os.path.basename(distro.initrd)
            if distro.breed == "redhat":
                append_line += " inst.ks=cdrom:/isolinux/%s.cfg" % descendant.name
            if distro.breed == "suse":
                append_line += " autoyast=file:///isolinux/%s.cfg install=cdrom:///" % descendant.name
                if "install" in data["kernel_options"]:
                    del data["kernel_options"]["install"]
            if distro.breed in ["ubuntu", "debian"]:
                append_line += " auto-install/enable=true preseed/file=/cdrom/isolinux/%s.cfg" % descendant.name

            # add remaining kernel_options to append_line
            append_line += self.add_remaining_kopts(data["kernel_options"])
            cfg.write(append_line)

            if descendant.COLLECTION_TYPE == 'profile':
                autoinstall_data = self.api.autoinstallgen.generate_autoinstall_for_profile(descendant.name)
            elif descendant.COLLECTION_TYPE == 'system':
                autoinstall_data = self.api.autoinstallgen.generate_autoinstall_for_system(descendant.name)

            if distro.breed == "redhat":
                cdregex = re.compile(r"^\s*url .*\n", re.IGNORECASE | re.MULTILINE)
                autoinstall_data = cdregex.sub("cdrom\n", autoinstall_data, count=1)

            if airgapped:
                descendant_repos = data['repos']
                for repo_name in descendant_repos:
                    repo_obj = self.api.find_repo(repo_name)
                    error_fmt = (descendant.COLLECTION_TYPE + " " + descendant.name + " refers to repo " + repo_name
                                 + ", which %%s; cannot build airgapped ISO")

                    if repo_obj is None:
                        utils.die(error_fmt % "does not exist")
                    if not repo_obj.mirror_locally:
                        utils.die(error_fmt % "is not configured for local mirroring")
                    # FIXME: don't hardcode
                    mirrordir = os.path.join(self.settings.webdir, "repo_mirror", repo_obj.name)
                    if not os.path.exists(mirrordir):
                        utils.die(error_fmt % "has a missing local mirror directory")

                    repo_names_to_copy[repo_obj.name] = mirrordir

                    # update the baseurl in autoinstall_data to use the cdrom copy of this repo
                    reporegex = re.compile(r"^(\s*repo --name=" + repo_obj.name + " --baseurl=).*", re.MULTILINE)
                    autoinstall_data = reporegex.sub(r"\1" + "file:///mnt/source/repo_mirror/" + repo_obj.name,
                                                     autoinstall_data)

                # rewrite any split-tree repos, such as in redhat, to use cdrom
                srcreporegex = re.compile(r"^(\s*repo --name=\S+ --baseurl=).*/cobbler/distro_mirror/" + distro.name
                                          + r"/?(.*)", re.MULTILINE)
                autoinstall_data = srcreporegex.sub(r"\1" + "file:///mnt/source" + r"\2", autoinstall_data)

            autoinstall_name = os.path.join(isolinuxdir, "%s.cfg" % descendant.name)
            with open(autoinstall_name, "w+") as autoinstall_file:
                autoinstall_file.write(autoinstall_data)

        self.logger.info("done writing config")
        cfg.write("\n")
        cfg.write("MENU END\n")
        cfg.close()

        if airgapped:
            # copy any repos found in profiles or systems to the iso build
            repodir = os.path.abspath(os.path.join(isolinuxdir, "..", "repo_mirror"))
            if not os.path.exists(repodir):
                os.makedirs(repodir)

            for repo_name in repo_names_to_copy:
                src = repo_names_to_copy[repo_name]
                dst = os.path.join(repodir, repo_name)
                self.logger.info(" - copying repo '" + repo_name + "' for airgapped ISO")

                ok = utils.rsync_files(src, dst, "--exclude=TRANS.TBL --exclude=cache/ --no-g",
                                       quiet=True)
                if not ok:
                    utils.die("rsync of repo '" + repo_name + "' failed")

        # copy distro files last, since they take the most time
        cmd = "rsync -rlptgu --exclude=boot.cat --exclude=TRANS.TBL --exclude=isolinux/ %s/ %s/../"\
              % (filesource, isolinuxdir)
        self.logger.info("- copying distro %s files (%s)" % (distname, cmd))
        rc = utils.subprocess_call(cmd, shell=True)
        if rc:
            utils.die("rsync of distro files failed")

    def run(self, iso=None, buildisodir=None, profiles=None, systems=None, distro=None,
            standalone: Optional[bool] = None, airgapped: Optional[bool] = None, source=None,
            exclude_dns: Optional[bool] = None, xorrisofs_opts: Optional[str] = None):
        """
        A

        :param iso: The name of the iso. Defaults to "autoinst.iso".
        :param buildisodir: This overwrites the directory from the settings in which the iso is built in.
        :param profiles:
        :param systems: Don't use that when building standalone isos.
        :param distro: (For standalone only)
        :param standalone: This means that no network connection is needed to install the generated iso.
        :param airgapped: This option implies standalone=True.
        :param source: If the iso should be offline available this is the path to the sources of the image.
        :param exclude_dns: Whether the repositories have to be locally available or the internet is reachable.
        :param xorrisofs_opts: xorrisofs options to include additionally.
        """

        if airgapped is True:
            standalone = True

        # the distro option is for stand-alone builds only
        if not standalone and distro is not None:
            utils.die("The --distro option should only be used when creating a standalone or airgapped ISO")
        # if building standalone, we only want --distro and --profiles (optional), systems are disallowed
        if standalone:
            if systems is not None:
                utils.die("When building a standalone ISO, use --distro and --profiles only, not --systems")
            elif distro is None:
                utils.die("When building a standalone ISO, you must specify a --distro")
            if source is not None and not os.path.exists(source):
                utils.die("The source specified (%s) does not exist" % source)

            # insure all profiles specified are children of the distro
            if profiles:
                which_profiles = self.filter_systems_or_profiles(profiles, 'profile')
                for profile in which_profiles:
                    if profile.distro != distro:
                        utils.die("When building a standalone ISO, all --profiles must be under --distro")

        # if iso is none, create it in . as "autoinst.iso"
        if iso is None:
            iso = "autoinst.iso"

        if buildisodir is None:
            buildisodir = self.settings.buildisodir
        else:
            if not os.path.isdir(buildisodir):
                utils.die("The --tempdir specified is not a directory")

            (buildisodir_head, buildisodir_tail) = os.path.split(os.path.normpath(buildisodir))
            if buildisodir_tail != "buildiso":
                buildisodir = os.path.join(buildisodir, "buildiso")

        self.logger.info("using/creating buildisodir: %s" % buildisodir)
        if not os.path.exists(buildisodir):
            os.makedirs(buildisodir)
        else:
            shutil.rmtree(buildisodir)
            os.makedirs(buildisodir)

        # if base of buildisodir does not exist, fail create all profiles unless filtered by "profiles"

        imagesdir = os.path.join(buildisodir, "images")
        isolinuxdir = os.path.join(buildisodir, "isolinux")

        self.logger.info("building tree for isolinux")
        if not os.path.exists(imagesdir):
            os.makedirs(imagesdir)
        if not os.path.exists(isolinuxdir):
            os.makedirs(isolinuxdir)

        self.logger.info("copying miscellaneous files")

        files_to_copy = ["isolinux.bin", "menu.c32", "chain.c32",
                         "ldlinux.c32", "libcom32.c32", "libutil.c32"]

        optional_files = ["ldlinux.c32", "libcom32.c32", "libutil.c32"]

        syslinux_folders = ["/usr/share/syslinux/",
                            "/usr/lib/syslinux/modules/bios/",
                            "/usr/lib/syslinux/",
                            "/usr/lib/ISOLINUX/"]

        # file_copy_success will be used to check for missing files
        file_copy_success = {f: False for f in files_to_copy if f not in optional_files}
        for syslinux_folder in syslinux_folders:
            if os.path.isdir(os.path.join(syslinux_folder)):
                for file_to_copy in files_to_copy:
                    source_file = os.path.join(syslinux_folder, file_to_copy)
                    if os.path.exists(source_file):
                        utils.copyfile(source_file, os.path.join(isolinuxdir, file_to_copy))
                        file_copy_success[file_to_copy] = True

        if False in file_copy_success.values():
            for k, v in file_copy_success:
                if not v:
                    self.logger.error("File not found: %s" % k)
            utils.die("Required file(s) not found. Please check your syslinux installation")

        if standalone or airgapped:
            self.generate_standalone_iso(imagesdir, isolinuxdir, distro, source, airgapped, profiles)
        else:
            self.generate_netboot_iso(imagesdir, isolinuxdir, profiles, systems, exclude_dns)

        if xorrisofs_opts is None:
            xorrisofs_opts = ""
        else:
            xorrisofs_opts = xorrisofs_opts.strip()

        # using xorrisofs instead of mkisofs nowadays, it is available everywhere...
        cmd = "xorrisofs -o %s %s -r -b isolinux/isolinux.bin -c isolinux/boot.cat" % (iso, xorrisofs_opts)
        cmd = cmd + " -no-emul-boot -boot-load-size 4"
        cmd = cmd + r" -boot-info-table -V Cobbler\ Install -R -J %s" % buildisodir

        rc = utils.subprocess_call(cmd, shell=True)
        if rc != 0:
            utils.die("xorrisofs failed")

        self.logger.info("ISO build complete")
        self.logger.info("You may wish to delete: %s" % buildisodir)
        self.logger.info("The output file is: %s" % iso)

# EOF



import distutils.sysconfig
import logging
import sys
import os
import time
import cobbler.templar
from cobbler.cexceptions import CX

plib = distutils.sysconfig.get_python_lib()
mod_path = "%s/cobbler" % plib
sys.path.insert(0, mod_path)
template_file = "/etc/cobbler/genders.template"
settings_file = "/etc/genders"

logger = logging.getLogger()


def register() -> str:
    """
    We should run anytime something inside of Cobbler changes.

    :return: Always ``/var/lib/cobbler/triggers/change/*``
    """
    return "/var/lib/cobbler/triggers/change/*"


def write_genders_file(config, profiles_genders, distros_genders, mgmtcls_genders):
    """
    Genders file is over-written when ``manage_genders`` is set in our settings.

    :param config: The config file to template with the data.
    :param profiles_genders: The profiles which should be included.
    :param distros_genders: The distros which should be included.
    :param mgmtcls_genders: The management classes which should be included.
    :raises OSError: Raised in case the template could not be read.
    """

    templar_inst = cobbler.templar.Templar(config)
    try:
        f2 = open(template_file, "r")
    except:
        raise OSError("error reading template: %s" % template_file)
    template_data = ""
    template_data = f2.read()
    f2.close()

    metadata = {
        "date": time.asctime(time.gmtime()),
        "profiles_genders": profiles_genders,
        "distros_genders": distros_genders,
        "mgmtcls_genders": mgmtcls_genders
    }

    templar_inst.render(template_data, metadata, settings_file)


def run(api, args) -> int:
    """
    Mandatory Cobbler trigger hook.

    :param api: The api to resolve information with.
    :param args: For this implementation unused.
    :return: ``0`` or ``1``, depending on the outcome of the operation.
    """
    # do not run if we are not enabled.
    if not api.settings().manage_genders:
        return 0

    profiles_genders = dict()
    distros_genders = dict()
    mgmtcls_genders = dict()

    # let's populate our dicts

    # TODO: the lists that are created here are strictly comma separated.
    # /etc/genders allows for host lists that are in the notation similar to: node00[01-07,08,09,70-71] at some point,
    # need to come up with code to generate these types of lists.

    # profiles
    for prof in api.profiles():
        # create the key
        profiles_genders[prof.name] = ""
        for system in api.find_system(profile=prof.name, return_list=True):
            profiles_genders[prof.name] += system.name + ","
        # remove a trailing comma
        profiles_genders[prof.name] = profiles_genders[prof.name][:-1]
        if profiles_genders[prof.name] == "":
            profiles_genders.pop(prof.name, None)

    # distros
    for dist in api.distros():
        # create the key
        distros_genders[dist.name] = ""
        for system in api.find_system(distro=dist.name, return_list=True):
            distros_genders[dist.name] += system.name + ","
        # remove a trailing comma
        distros_genders[dist.name] = distros_genders[dist.name][:-1]
        if distros_genders[dist.name] == "":
            distros_genders.pop(dist.name, None)

    # mgmtclasses
    for mgmtcls in api.mgmtclasses():
        # create the key
        mgmtcls_genders[mgmtcls.name] = ""
        for system in api.find_system(mgmt_classes=mgmtcls.name, return_list=True):
            mgmtcls_genders[mgmtcls.name] += system.name + ","
        # remove a trailing comma
        mgmtcls_genders[mgmtcls.name] = mgmtcls_genders[mgmtcls.name][:-1]
        if mgmtcls_genders[mgmtcls.name] == "":
            mgmtcls_genders.pop(mgmtcls.name, None)
    # The file doesn't exist and for some reason the template engine won't create it, so spit out an error and tell the
    # user what to do.
    if not os.path.isfile(settings_file):
        logger.info("Error: " + settings_file + " does not exist.")
        logger.info("Please run: touch " + settings_file + " as root and try again.")
        return 1

    write_genders_file(api._collection_mgr, profiles_genders, distros_genders, mgmtcls_genders)
    return 0



from xmlrpc.client import Fault

import pytest


@pytest.mark.usefixtures("cobbler_xmlrpc_base")
class TestSystem:
    """
    Test remote calls related to systems.
    """

    def test_get_systems(self, remote, token):
        """
        Test: get systems
        """
        # Arrange --> Nothing to arrange

        # Act
        result = remote.get_systems(token)

        # Assert
        assert result == []

    @pytest.mark.usefixtures("create_testdistro", "create_testmenu", "create_testprofile", "remove_testdistro",
                             "remove_testmenu", "remove_testprofile", "remove_testsystem")
    @pytest.mark.parametrize("field_name,field_value", [
        ("comment", "test comment"),
        ("enable_ipxe", True),
        ("enable_ipxe", False),
        ("kernel_options", "a=1 b=2 c=3 c=4 c=5 d e"),
        ("kernel_options_post", "a=1 b=2 c=3 c=4 c=5 d e"),
        ("autoinstall", "test.ks"),
        ("autoinstall", "test.xml"),
        ("autoinstall", "test.seed"),
        ("autoinstall_meta", "a=1 b=2 c=3 c=4 c=5 d e"),
        ("mgmt_classes", "one two three"),
        ("mgmt_parameters", "<<inherit>>"),
        ("name", "testsystem0"),
        ("netboot_enabled", True),
        ("netboot_enabled", False),
        ("owners", "user1 user2 user3"),
        ("profile", "testprofile0"),
        ("repos_enabled", True),
        ("repos_enabled", False),
        ("status", "development"),
        ("status", "testing"),
        ("status", "acceptance"),
        ("status", "production"),
        ("proxy", "testproxy"),
        ("server", "1.1.1.1"),
        # ("boot_loaders", "pxe ipxe grub"), FIXME: This raises currently but it did not in the past
        ("virt_auto_boot", True),
        ("virt_auto_boot", False),
        ("virt_auto_boot", "yes"),
        ("virt_auto_boot", "no"),
        ("virt_cpus", "<<inherit>>"),
        ("virt_cpus", 1),
        ("virt_cpus", 2),
        ("virt_cpus", "<<inherit>>"),
        ("virt_file_size", "<<inherit>>"),
        ("virt_file_size", 5),
        ("virt_file_size", 10),
        ("virt_disk_driver", "<<inherit>>"),
        ("virt_disk_driver", "raw"),
        ("virt_disk_driver", "qcow2"),
        ("virt_disk_driver", "vdmk"),
        ("virt_ram", "<<inherit>>"),
        ("virt_ram", 256),
        ("virt_ram", 1024),
        ("virt_type", "<<inherit>>"),
        ("virt_type", "xenpv"),
        ("virt_type", "xenfv"),
        ("virt_type", "qemu"),
        ("virt_type", "kvm"),
        ("virt_type", "vmware"),
        ("virt_type", "openvz"),
        ("virt_path", "<<inherit>>"),
        ("virt_path", "/path/to/test"),
        ("virt_pxe_boot", True),
        ("virt_pxe_boot", False),
        ("power_type", "ipmilanplus"),
        ("power_address", "127.0.0.1"),
        ("power_id", "pmachine:lpar1"),
        ("power_pass", "pass"),
        ("power_user", "user")
    ])
    def test_create_system_positive(self, remote, token, template_files, field_name, field_value):
        """
        Test: create/edit a system object
        """
        # Arrange
        system = remote.new_system(token)
        remote.modify_system(system, "name", "testsystem0", token)
        remote.modify_system(system, "profile", "testprofile0", token)

        # Act
        result = remote.modify_system(system, field_name, field_value, token)

        # Assert
        assert result

    @pytest.mark.usefixtures("create_testdistro", "create_testmenu", "create_testprofile", "remove_testdistro",
                             "remove_testmenu", "remove_testprofile", "remove_testsystem")
    @pytest.mark.parametrize("field_name,field_value", [
        ("autoinstall", "/path/to/bad/autoinstall"),
        ("mgmt_parameters", "badyaml"),
        ("profile", "badprofile"),
        ("boot_loaders", "badloader"),
        ("virt_cpus", "a"),
        ("virt_file_size", "a"),
        ("virt_ram", "a"),
        ("virt_type", "bad"),
        ("power_type", "bla")
    ])
    def test_create_system_negative(self, remote, token, field_name, field_value):
        """
        Test: create/edit a system object
        """
        # Arrange
        system = remote.new_system(token)
        remote.modify_system(system, "name", "testsystem0", token)
        remote.modify_system(system, "profile", "testprofile0", token)

        # Act & Assert
        try:
            remote.modify_system(system, field_name, field_value, token)
        except Fault:
            assert True
        else:
            pytest.fail("Bad field did not raise an exception!")

    def test_find_system(self, remote, token):
        """
        Test: find a system object
        """

        # Arrange --> Nothing to arrange

        # Act
        result = remote.find_system({"name": "notexisting"}, token)

        # Assert --> A not exiting system returns an empty list
        assert result == []

    @pytest.mark.usefixtures("create_testdistro", "create_testmenu", "create_testprofile", "create_testsystem",
                             "remove_testdistro",
                             "remove_testmenu", "remove_testprofile", "remove_testsystem")
    def test_copy_system(self, remote, token):
        """
        Test: copy a system object
        """

        # Arrange
        system = remote.get_item_handle("system", "testsystem0", token)

        # Act
        result = remote.copy_system(system, "testsystemcopy", token)

        # Assert
        assert result

        # Cleanup
        remote.remove_system("testsytemcopy", token)

    @pytest.mark.usefixtures("create_testdistro", "create_testmenu", "create_testprofile", "create_testsystem",
                             "remove_testdistro",
                             "remove_testmenu", "remove_testprofile", "remove_testsystem")
    def test_rename_system(self, remote, token):
        """
        Test: rename a system object
        """

        # Arrange --> Done in fixtures also.
        system = remote.get_item_handle("system", "testsystem0", token)

        # Act
        result = remote.rename_system(system, "testsystem1", token)

        # Assert
        assert result

        # Cleanup
        remote.remove_system("testsystem1", token)

    def test_remove_system(self, remote, token):
        """
        Test: remove a system object
        """

        # Arrange
        # TODO: Verify why the test passes without the fixture for creating the system!

        # Act
        result = remote.remove_system("testsystem0", token)

        # Assert
        assert result

    def test_get_repo_config_for_system(self, remote):
        """
        Test: get repository configuration of a system
        """

        # Arrange --> There is nothing to be arranged

        # Act
        result = remote.get_repo_config_for_system("testprofile0")

        # Assert --> Let the test pass if the call is okay.
        assert True



# -*- coding: utf-8 -*-
#
# Adrian Brzezinski <adrbxx@gmail.com>
# License: GPLv2+
#
# Replace (or remove) records in DNS zone for systems created (or removed) by Cobbler
#

# DNS toolkit for Python
#   - python-dnspython (Debian)
#   - python-dns (RH/CentOS)

import dns.query
import dns.tsigkeyring
import dns.update
import dns.resolver
import time

from cobbler.cexceptions import CX

logf = None


def nslog(msg):
    """
    Log a message to the logger.

    :param msg: The message to log.
    """
    if logf is not None:
        logf.write(msg)


def register() -> str:
    """
    This method is the obligatory Cobbler registration hook.

    :return: The trigger name or an empty string.
    """
    if __name__ == "cobbler.modules.nsupdate_add_system_post":
        return "/var/lib/cobbler/triggers/add/system/post/*"
    elif __name__ == "cobbler.modules.nsupdate_delete_system_pre":
        return "/var/lib/cobbler/triggers/delete/system/pre/*"
    else:
        return ''


def run(api, args):
    """
    This method executes the trigger, meaning in this case that it updates the dns configuration.

    :param api: The api to read metadata from.
    :param args: Metadata to log.
    :return: "0" on success or a skipped task. If the task failed or problems occurred then an exception is raised.
    """
    global logf

    action = None
    if __name__ == "cobbler.modules.nsupdate_add_system_post":
        action = "replace"
    elif __name__ == "cobbler.modules.nsupdate_delete_system_pre":
        action = "delete"
    else:
        return 0

    settings = api.settings()

    if not settings.nsupdate_enabled:
        return 0

    # read our settings
    if str(settings.nsupdate_log) is not None:
        logf = open(str(settings.nsupdate_log), "a+")
        nslog(">> starting %s %s\n" % (__name__, args))

    if str(settings.nsupdate_tsig_key) is not None:
        keyring = dns.tsigkeyring.from_text({
            str(settings.nsupdate_tsig_key[0]): str(settings.nsupdate_tsig_key[1])
        })
    else:
        keyring = None

    if str(settings.nsupdate_tsig_algorithm) is not None:
        keyring_algo = str(settings.nsupdate_tsig_algorithm)
    else:
        keyring_algo = "HMAC-MD5.SIG-ALG.REG.INT"
    # nslog( " algo %s, key %s : %s \n" % (keyring_algo,str(settings.nsupdate_tsig_key[0]),
    #                                      str(settings.nsupdate_tsig_key[1])) )

    # get information about this system
    system = api.find_system(args[0])

    # process all interfaces and perform dynamic update for those with --dns-name
    for (name, interface) in list(system.interfaces.items()):
        host = interface["dns_name"]
        host_ip = interface["ip_address"]

        if not system.is_management_supported(cidr_ok=False):
            continue
        if not host:
            continue
        if host.find(".") == -1:
            continue

        domain = ".".join(host.split(".")[1:])    # get domain from host name
        host = host.split(".")[0]                   # strip domain

        nslog("processing interface %s : %s\n" % (name, interface))
        nslog("lookup for '%s' domain master nameserver... " % domain)

        # get master nameserver ip address
        answers = dns.resolver.query(domain + '.', dns.rdatatype.SOA)
        soa_mname = answers[0].mname
        soa_mname_ip = None

        for rrset in answers.response.additional:
            if rrset.name == soa_mname:
                soa_mname_ip = str(rrset.items[0].address)

        if soa_mname_ip is None:
            ip = dns.resolver.query(soa_mname, "A")
            for answer in ip:
                soa_mname_ip = answer.to_text()

        nslog("%s [%s]\n" % (soa_mname, soa_mname_ip))
        nslog("%s dns record for %s.%s [%s] .. " % (action, host, domain, host_ip))

        # try to update zone with new record
        update = dns.update.Update(domain + '.', keyring=keyring, keyalgorithm=keyring_algo)

        if action == "replace":
            update.replace(host, 3600, dns.rdatatype.A, host_ip)
            update.replace(host, 3600, dns.rdatatype.TXT, '"cobbler (date: %s)"' % (time.strftime("%c")))
        else:
            update.delete(host, dns.rdatatype.A, host_ip)
            update.delete(host, dns.rdatatype.TXT)

        try:
            response = dns.query.tcp(update, soa_mname_ip)
            rcode_txt = dns.rcode.to_text(response.rcode())
        except dns.tsig.PeerBadKey:
            nslog("failed (refused key)\n>> done\n")
            logf.close()

            raise CX("nsupdate failed, server '%s' refusing our key" % soa_mname)

        nslog('response code: %s\n' % rcode_txt)

        # notice user about update failure
        if response.rcode() != dns.rcode.NOERROR:
            nslog('>> done\n')
            logf.close()

            raise CX("nsupdate failed (response: %s, name: %s.%s, ip %s, name server %s)"
                     % (rcode_txt, host, domain, host_ip, soa_mname))

    nslog('>> done\n')
    logf.close()
    return 0



import os

import pytest

from cobbler.api import CobblerAPI
from cobbler.configgen import ConfigGen
from cobbler.items.distro import Distro
from cobbler.items.profile import Profile
from cobbler.items.system import System


# TODO: If the action items of the configgen class are done then the tests need to test the robustness of the class.


@pytest.fixture
def create_testbed(create_kernel_initrd):
    def _create_testbed() -> CobblerAPI:
        folder = create_kernel_initrd("vmlinux", "initrd.img")
        test_api = CobblerAPI()
        test_distro = Distro(test_api)
        test_distro.name = "test_configgen_distro"
        test_distro.kernel = os.path.join(folder, "vmlinux")
        test_distro.initrd = os.path.join(folder, "initrd.img")
        test_api.add_distro(test_distro)
        test_profile = Profile(test_api)
        test_profile.name = "test_configgen_profile"
        test_profile.distro = "test_configgen_distro"
        test_api.add_profile(test_profile)
        test_system = System(test_api)
        test_system.name = "test_configgen_system"
        test_system.profile = "test_configgen_profile"
        test_system.hostname = "testhost.test.de"
        test_system.autoinstall_meta = {"test": "teststring"}
        test_api.add_system(test_system)
        return test_api

    return _create_testbed


def test_object_value_error():
    # Arrange
    test_api = CobblerAPI()

    # Act & Assert
    with pytest.raises(ValueError):
        ConfigGen(test_api, "nonexistant")


def test_object_creation(create_testbed):
    # Arrange
    test_api = create_testbed()
    # FIXME: Arrange distro, profile and system

    # Act
    test_configgen = ConfigGen(test_api, "testhost.test.de")

    # Assert
    assert isinstance(test_configgen, ConfigGen)


def test_resolve_resource_var(create_testbed):
    # Arrange
    test_api = create_testbed()
    # FIXME: Arrange distro, profile and system, package and file
    config_gen = ConfigGen(test_api, "testhost.test.de")

    # Act
    result = config_gen.resolve_resource_var("Hello $test !")

    # Assert
    assert isinstance(result, str)
    assert result == "Hello teststring !"


def test_get_cobbler_resource(create_testbed):
    # Arrange
    test_api = create_testbed()
    # FIXME: Arrange distro, profile and system, package and file
    config_gen = ConfigGen(test_api, "testhost.test.de")

    # Act
    result = config_gen.get_cobbler_resource("")

    # Assert
    assert isinstance(result, (list, str, dict))


def test_get_config_data(create_testbed):
    # Arrange
    test_api = create_testbed()
    # FIXME: Arrange distro, profile and system, package and file
    config_gen = ConfigGen(test_api, "testhost.test.de")

    # Act
    result = config_gen.gen_config_data()

    # Assert
    assert isinstance(result, dict)


def test_get_config_data_for_koan(create_testbed):
    # Arrange
    test_api = create_testbed()
    # FIXME: Arrange distro, profile and system, package and file
    config_gen = ConfigGen(test_api, "testhost.test.de")

    # Act
    result = config_gen.gen_config_data_for_koan()

    # Assert
    assert isinstance(result, str)



"""
Migration from V3.2.1 to V3.3.0
"""
# SPDX-License-Identifier: GPL-2.0-or-later
# SPDX-FileCopyrightText: 2021 Dominik Gedon <dgedon@suse.de>
# SPDX-FileCopyrightText: 2021 Enno Gotthold <egotthold@suse.de>
# SPDX-FileCopyrightText: Copyright SUSE LLC


from schema import Optional, Schema, SchemaError

from cobbler.settings.migrations import helper

schema = Schema({
    "auto_migrate_settings": bool,
    "allow_duplicate_hostnames": bool,
    "allow_duplicate_ips": bool,
    "allow_duplicate_macs": bool,
    "allow_dynamic_settings": bool,
    "always_write_dhcp_entries": bool,
    "anamon_enabled": bool,
    "auth_token_expiration": int,
    "authn_pam_service": str,
    "autoinstall_snippets_dir": str,
    "autoinstall_templates_dir": str,
    "bind_chroot_path": str,
    "bind_zonefile_path": str,
    "bind_master": str,
    "boot_loader_conf_template_dir": str,
    Optional("bootloaders_dir", default="/var/lib/cobbler/loaders"): str,
    Optional("bootloaders_formats", default={
        'aarch64': {'binary_name': 'grubaa64.efi'},
        'arm': {'binary_name': 'bootarm.efi'},
        'arm64-efi': {'binary_name': 'grubaa64.efi', 'extra_modules': ['efinet']},
        'i386': {'binary_name': 'bootia32.efi'},
        'i386-pc-pxe': {'binary_name': 'grub.0', 'mod_dir': 'i386-pc',
                        'extra_modules': ['chain', 'pxe', 'biosdisk']},
        'i686': {'binary_name': 'bootia32.efi'},
        'IA64': {'binary_name': 'bootia64.efi'},
        'powerpc-ieee1275': {'binary_name': 'grub.ppc64le', 'extra_modules': ['net', 'ofnet']},
        'x86_64-efi': {'binary_name': 'grubx86.efi', 'extra_modules': ['chain', 'efinet']}}
             ): dict,
    Optional("bootloaders_modules", default=[
        'btrfs', 'ext2', 'xfs', 'jfs', 'reiserfs', 'all_video', 'boot',
        'cat', 'configfile', 'echo', 'fat', 'font', 'gfxmenu', 'gfxterm',
        'gzio', 'halt', 'iso9660', 'jpeg', 'linux', 'loadenv', 'minicmd',
        'normal', 'part_apple', 'part_gpt', 'part_msdos', 'password_pbkdf2',
        'png', 'reboot', 'search', 'search_fs_file', 'search_fs_uuid',
        'search_label', 'sleep', 'test', 'true', 'video', 'mdraid09',
        'mdraid1x', 'lvm', 'serial', 'regexp', 'tr', 'tftp', 'http', 'luks',
        'gcry_rijndael', 'gcry_sha1', 'gcry_sha256'
    ]): list,
    Optional("syslinux_dir", default="/usr/share/syslinux"): str,
    Optional("grub2_mod_dir", default="/usr/share/grub"): str,
    Optional("grubconfig_dir", default="/var/lib/cobbler/grub_config"): str,
    "build_reporting_enabled": bool,
    "build_reporting_email": [str],
    "build_reporting_ignorelist": [str],
    "build_reporting_sender": str,
    "build_reporting_smtp_server": str,
    "build_reporting_subject": str,
    Optional("buildisodir", default="/var/cache/cobbler/buildiso"): str,
    "cheetah_import_whitelist": [str],
    "client_use_https": bool,
    "client_use_localhost": bool,
    Optional("cobbler_master", default=""): str,
    Optional("convert_server_to_ip", default=False): bool,
    "createrepo_flags": str,
    "autoinstall": str,
    "default_name_servers": [str],
    "default_name_servers_search": [str],
    "default_ownership": [str],
    "default_password_crypted": str,
    "default_template_type": str,
    "default_virt_bridge": str,
    Optional("default_virt_disk_driver", default="raw"): str,
    "default_virt_file_size": int,
    "default_virt_ram": int,
    "default_virt_type": str,
    "enable_ipxe": bool,
    "enable_menu": bool,
    "http_port": int,
    "include": [str],
    Optional("iso_template_dir", default="/etc/cobbler/iso"): str,
    Optional("jinja2_includedir", default="/var/lib/cobbler/jinja2"): str,
    "kernel_options": dict,
    "ldap_anonymous_bind": bool,
    "ldap_base_dn": str,
    "ldap_port": int,
    "ldap_search_bind_dn": str,
    "ldap_search_passwd": str,
    "ldap_search_prefix": str,
    "ldap_server": str,
    "ldap_tls": bool,
    "ldap_tls_cacertfile": str,
    "ldap_tls_certfile": str,
    "ldap_tls_keyfile": str,
    Optional("bind_manage_ipmi", default=False): bool,
    # TODO: Remove following line
    "manage_dhcp": bool,
    "manage_dhcp_v4": bool,
    "manage_dhcp_v6": bool,
    "manage_dns": bool,
    "manage_forward_zones": [str],
    "manage_reverse_zones": [str],
    Optional("manage_genders", False): bool,
    "manage_rsync": bool,
    "manage_tftpd": bool,
    "mgmt_classes": [str],
    # TODO: Validate Subdict
    "mgmt_parameters": dict,
    "next_server_v4": str,
    "next_server_v6": str,
    Optional("nsupdate_enabled", False): bool,
    Optional("nsupdate_log", default="/var/log/cobbler/nsupdate.log"): str,
    Optional("nsupdate_tsig_algorithm", default="hmac-sha512"): str,
    Optional("nsupdate_tsig_key", default=[]): [str],
    "power_management_default_type": str,
    "proxy_url_ext": str,
    "proxy_url_int": str,
    "puppet_auto_setup": bool,
    Optional("puppet_parameterized_classes", default=True): bool,
    Optional("puppet_server", default="puppet"): str,
    Optional("puppet_version", default=2): int,
    "puppetca_path": str,
    "pxe_just_once": bool,
    "nopxe_with_triggers": bool,
    "redhat_management_permissive": bool,
    "redhat_management_server": str,
    "redhat_management_key": str,
    "register_new_installs": bool,
    "remove_old_puppet_certs_automatically": bool,
    "replicate_repo_rsync_options": str,
    "replicate_rsync_options": str,
    "reposync_flags": str,
    "reposync_rsync_flags": str,
    "restart_dhcp": bool,
    "restart_dns": bool,
    "run_install_triggers": bool,
    "scm_track_enabled": bool,
    "scm_track_mode": str,
    "scm_track_author": str,
    "scm_push_script": str,
    "serializer_pretty_json": bool,
    "server": str,
    "sign_puppet_certs_automatically": bool,
    Optional("signature_path", default="/var/lib/cobbler/distro_signatures.json"): str,
    Optional("signature_url", default="https://cobbler.github.io/signatures/3.0.x/latest.json"): str,
    "tftpboot_location": str,
    "virt_auto_boot": bool,
    "webdir": str,
    "webdir_whitelist": [str],
    "xmlrpc_port": int,
    "yum_distro_priority": int,
    "yum_post_install_mirror": bool,
    "yumdownloader_flags": str,
    Optional("windows_enabled", default=False): bool,
    Optional("windows_template_dir", default="/etc/cobbler/windows"): str,
    Optional("samba_distro_share", default="DISTRO"): str,
}, ignore_extra_keys=False)


def validate(settings: dict) -> bool:
    """
    Checks that a given settings dict is valid according to the reference V3.3.0 schema ``schema``.

    :param settings: The settings dict to validate.
    :return: True if valid settings dict otherwise False.
    """
    try:
        schema.validate(settings)
    except SchemaError:
        return False
    return True


def normalize(settings: dict) -> dict:
    """
    If data in ``settings`` is valid the validated data is returned.

    :param settings: The settings dict to validate.
    :return: The validated dict.
    """
    return schema.validate(settings)


def migrate(settings: dict) -> dict:
    """
    Migration of the settings ``settings`` to version V3.3.0 settings

    :param settings: The settings dict to migrate
    :return: The migrated dict
    """

    # migrate gpxe -> ipxe
    if "enable_gpxe" in settings:
        gpxe = helper.key_get("enable_gpxe", settings)
        helper.key_rename(gpxe, "enable_ipxe", settings)

    # rename keys and update their value
    old_setting = helper.Setting("default_autoinstall", "/var/lib/cobbler/autoinstall_templates/default.ks")
    new_setting = helper.Setting("autoinstall", "default.ks")
    helper.key_rename(old_setting, "autoinstall", settings)
    helper.key_set_value(new_setting, settings)

    old_setting = helper.Setting("next_server", "127.0.0.1")
    new_setting = helper.Setting("next_server_v4", "127.0.0.1")
    helper.key_rename(old_setting, "next_server_v4", settings)
    helper.key_set_value(new_setting, settings)

    # add missing keys
    # name - value pairs
    missing_keys = {'auto_migrate_settings': True,
                    'bind_zonefile_path': "@@bind_zonefiles@@",
                    'bootloaders_formats': {
                        "aarch64": {"binary_name": "grubaa64.efi"},
                        "arm": {"binary_name": "bootarm.efi"},
                        "arm64-efi": {"binary_name": "grubaa64.efi", "extra_modules": ["efinet"]},
                        "i386": {"binary_name": "bootia32.efi"},
                        "i386-pc-pxe": {
                            "binary_name": "grub.0",
                            "mod_dir": "i386-pc",
                            "extra_modules": ["chain", "pxe", "biosdisk"]},
                        "i686": {"binary_name": "bootia32.efi"},
                        "IA64": {"binary_name": "bootia64.efi"},
                        "powerpc-ieee1275": {"binary_name": "grub.ppc64le", "extra_modules": ["net", "ofnet"], },
                        "x86_64-efi": {"binary_name": "grubx86.efi", "extra_modules": ["chain", "efinet"]}
                    },
                    'bootloaders_modules': ["btrfs", "ext2", "xfs", "jfs", "reiserfs",
                                            "all_video", "boot", "cat", "configfile", "echo", "fat", "font", "gfxmenu",
                                            "gfxterm", "gzio", "halt", "iso9660", "jpeg", "linux", "loadenv", "minicmd",
                                            "normal", "part_apple", "part_gpt", "part_msdos", "password_pbkdf2", "png",
                                            "reboot", "search", "search_fs_file", "search_fs_uuid", "search_label",
                                            "sleep", "test", "true", "video", "mdraid09", "mdraid1x", "lvm", "serial",
                                            "regexp", "tr", "tftp", "http", "luks", "gcry_rijndael", "gcry_sha1",
                                            "gcry_sha256"],
                    'grub2_mod_dir': "/usr/share/grub2",
                    'manage_dhcp_v4': False,
                    'manage_dhcp_v6': False,
                    'next_server_v6': "::1",
                    'syslinux_dir': "/usr/share/syslinux"}
    for (key, value) in missing_keys.items():
        new_setting = helper.Setting(key, value)
        helper.key_add(new_setting, settings)

    # delete removed keys
    helper.key_delete("cache_enabled", settings)

    if not validate(settings):
        raise SchemaError("V3.3.0: Schema error while validating")
    return normalize(settings)



"""
Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import uuid
from typing import Optional, Union

from cobbler import autoinstall_manager
from cobbler.items import item
from cobbler import utils, validate, enums
from cobbler.cexceptions import CX
from cobbler.items.distro import Distro


class Profile(item.Item):
    """
    A Cobbler profile object.
    """

    TYPE_NAME = "profile"
    COLLECTION_TYPE = "profile"

    def __init__(self, api, *args, **kwargs):
        """

        :param api: The Cobbler API object which is used for resolving information.
        :param args:
        :param kwargs:
        """
        super().__init__(api, *args, **kwargs)
        self._template_files = {}
        self._autoinstall = enums.VALUE_INHERITED
        self._boot_loaders: Union[list, str] = enums.VALUE_INHERITED
        self._dhcp_tag = ""
        self._distro = ""
        self._enable_ipxe = api.settings().enable_ipxe
        self._enable_menu = api.settings().enable_menu
        self._name_servers = api.settings().default_name_servers
        self._name_servers_search = api.settings().default_name_servers_search
        self._next_server_v4 = api.settings().next_server_v4
        self._next_server_v6 = api.settings().next_server_v6
        self._filename = ""
        self._proxy = enums.VALUE_INHERITED
        self._redhat_management_key = enums.VALUE_INHERITED
        self._repos = []
        self._server = enums.VALUE_INHERITED
        self._menu = ""
        # FIXME: The virt_* attributes don't support inheritance so far
        self._virt_auto_boot = api.settings().virt_auto_boot
        self._virt_bridge = api.settings().default_virt_bridge
        self._virt_cpus: Union[int, str] = 1
        self._virt_disk_driver = enums.VirtDiskDrivers.RAW
        self._virt_file_size = 0
        self._virt_path = ""
        self._virt_ram = api.settings().default_virt_ram
        self._virt_type = enums.VirtType.AUTO

        # Overwrite defaults from item.py
        self._boot_files = enums.VALUE_INHERITED
        self._fetchable_files = enums.VALUE_INHERITED
        self._autoinstall_meta = enums.VALUE_INHERITED
        self._kernel_options = enums.VALUE_INHERITED
        self._kernel_options_post = enums.VALUE_INHERITED
        self._mgmt_classes = enums.VALUE_INHERITED
        self._mgmt_parameters = enums.VALUE_INHERITED
        self._mgmt_classes = enums.VALUE_INHERITED

        # Use setters to validate settings
        self.virt_disk_driver = api.settings().default_virt_disk_driver
        self.virt_type = api.settings().default_virt_type

    def __getattr__(self, name):
        if name == "kickstart":
            return self.autoinstall
        elif name == "ks_meta":
            return self.autoinstall_meta
        raise AttributeError("Attribute \"%s\" did not exist on object type Profile." % name)

    #
    # override some base class methods first (item.Item)
    #

    def make_clone(self):
        """
        Clone this file object. Please manually adjust all value yourself to make the cloned object unique.

        :return: The cloned instance of this object.
        """
        _dict = self.to_dict()
        cloned = Profile(self.api)
        cloned.from_dict(_dict)
        cloned.uid = uuid.uuid4().hex
        return cloned

    def check_if_valid(self):
        """
        Check if the profile is valid. This checks for an existing name and a distro as a conceptual parent.

        :raises CX: In case the distro or name is not present.
        """
        # name validation
        if not self.name:
            raise CX("Name is required")

        # distro validation
        distro = self.get_conceptual_parent()
        if distro is None:
            raise CX("Error with profile %s - distro is required" % self.name)

    def from_dict(self, dictionary: dict):
        """
        Initializes the object with attributes from the dictionary.

        :param dictionary: The dictionary with values.
        """
        if "name" in dictionary:
            self.name = dictionary["name"]
        if "parent" in dictionary:
            self.parent = dictionary["parent"]
        if "distro" in dictionary:
            self.distro = dictionary["distro"]
        self._remove_depreacted_dict_keys(dictionary)
        super().from_dict(dictionary)

    #
    # specific methods for item.Profile
    #

    @property
    def parent(self) -> Optional[item.Item]:
        r"""
        Instead of a ``--distro``, set the parent of this object to another profile and use the values from the parent
        instead of this one where the values for this profile aren't filled in, and blend them together where they
        are dictionaries. Basically this enables profile inheritance. To use this, the object MUST have been
        constructed with ``is_subobject=True`` or the default values for everything will be screwed up and this will
        likely NOT work. So, API users -- make sure you pass ``is_subobject=True`` into the constructor when using this.

        Return object next highest up the tree. If this property is not set it falls back to the value of the
        ``distro``. In case neither distro nor parent is set, it returns None (which would make the profile invalid).

        :getter: The parent object which can be either another profile, a distro or None in case the object could not be
                 resolved.
        :setter: The name of the parent object. Might throw a ``CX`` in case the object could not be found.
        """
        if not self._parent:
            parent = self.distro
            if parent is None:
                return None
            return parent
        else:
            result = self.api.profiles().find(name=self._parent)
        return result

    @parent.setter
    def parent(self, parent: str):
        r"""
        Setter for the ``parent`` property.

        :param parent: The name of the parent object.
        :raises CX: In case self parentage is found or the profile given could not be found.
        """
        old_parent = self.parent
        if isinstance(old_parent, item.Item) and self.name in old_parent.children:
            old_parent.children.remove(self.name)
            self.api.serialize()
        if not parent:
            self._parent = ''
            return
        if parent == self.name:
            # check must be done in two places as setting parent could be called before/after setting name...
            raise CX("self parentage is weird")
        found = self.api.profiles().find(name=parent)
        if found is None:
            raise CX("profile %s not found, inheritance not possible" % parent)
        self._parent = parent
        self.depth = found.depth + 1
        new_parent = self.parent
        if isinstance(new_parent, item.Item) and self.name not in new_parent.children:
            new_parent.children.append(self.name)
            self.api.serialize()

    @property
    def arch(self):
        """
        This represents the architecture of a profile. It is read only.

        :getter: ``None`` or the parent architecture.
        """
        # FIXME: This looks so wrong. It cries: Please open a bug for me!
        parent = self.parent
        if parent:
            return parent.arch
        return None

    @property
    def distro(self):
        """
        The parent distro of a profile. This is not representing the Distro but the id of it.

        This is a required property, if saved to the disk, with the exception if this is a subprofile.

        :return: The distro object or None.
        """
        if not self._distro:
            return None
        return self.api.distros().find(name=self._distro)

    @distro.setter
    def distro(self, distro_name: str):
        """
        Sets the distro. This must be the name of an existing Distro object in the Distros collection.

        :param distro_name: The name of the distro.
        """
        if not isinstance(distro_name, str):
            raise TypeError("distro_name needs to be of type str")
        if not distro_name:
            self._distro = ""
            return
        distro = self.api.distros().find(name=distro_name)
        if distro is None:
            raise ValueError("distribution not found")
        old_parent = self.parent
        if isinstance(old_parent, item.Item) and self.name in old_parent.children:
            old_parent.children.remove(self.name)
        self._distro = distro_name
        self.depth = distro.depth + 1    # reset depth if previously a subprofile and now top-level
        if self.name not in distro.children:
            distro.children.append(self.name)
            self.api.serialize()

    @property
    def name_servers(self) -> list:
        """
        Represents the list of nameservers to set for the profile.

        :getter: The nameservers.
        :setter: Comma delimited ``str`` or list with the nameservers.
        """
        return self._resolve("name_servers")

    @name_servers.setter
    def name_servers(self, data: list):
        """
        Set the DNS servers.

        :param data: string or list of nameservers
        """
        self._name_servers = validate.name_servers(data)

    @property
    def name_servers_search(self) -> list:
        """
        Represents the list of DNS search paths.

        :getter: The list of DNS search paths.
        :setter: Comma delimited ``str`` or list with the nameservers search paths.
        """
        return self._resolve("name_servers_search")

    @name_servers_search.setter
    def name_servers_search(self, data: list):
        """
        Set the DNS search paths.

        :param data: string or list of search domains
        """
        self._name_servers_search = validate.name_servers_search(data)

    @property
    def proxy(self) -> str:
        """
        Override the default external proxy which is used for accessing the internet.

        :getter: Returns the default one or the specific one for this repository.
        :setter: May raise a ``TypeError`` in case the wrong value is given.
        """
        return self._resolve("proxy_url_int")

    @proxy.setter
    def proxy(self, proxy: str):
        """
        Setter for the proxy setting of the repository.

        :param proxy: The new proxy which will be used for the repository.
        :raises TypeError: In case the new value is not of type ``str``.
        """
        if not isinstance(proxy, str):
            raise TypeError("Field proxy of object profile needs to be of type str!")
        self._proxy = proxy

    @property
    def enable_ipxe(self) -> bool:
        r"""
        Sets whether or not the profile will use iPXE for booting.

        :getter: If set to inherit then this returns the parent value, otherwise it returns the real value.
        :setter: May throw a ``TypeError`` in case the new value cannot be cast to ``bool``.
        """
        return self._resolve("enable_ipxe")

    @enable_ipxe.setter
    def enable_ipxe(self, enable_ipxe: bool):
        r"""
        Setter for the ``enable_ipxe`` property.

        :param enable_ipxe: New boolean value for enabling iPXE.
        :raises TypeError: In case after the conversion, the new value is not of type ``bool``.
        """
        enable_ipxe = utils.input_boolean(enable_ipxe)
        if not isinstance(enable_ipxe, bool):
            raise TypeError("enable_ipxe needs to be of type bool")
        self._enable_ipxe = enable_ipxe

    @property
    def enable_menu(self) -> bool:
        """
        Sets whether or not the profile will be listed in the default PXE boot menu. This is pretty forgiving for
        YAML's sake.

        :getter: The value resolved from the defaults or the value specific to the profile.
        :setter: May raise a ``TypeError`` in case the boolean could not be converted.
        """
        return self._resolve("enable_menu")

    @enable_menu.setter
    def enable_menu(self, enable_menu: bool):
        """
        Setter for the ``enable_menu`` property.

        :param enable_menu: New boolean value for enabling the menu.
        :raises TypeError: In case the boolean could not be converted successfully.
        """
        enable_menu = utils.input_boolean(enable_menu)
        if not isinstance(enable_menu, bool):
            raise TypeError("enable_menu needs to be of type bool")
        self._enable_menu = enable_menu

    @property
    def dhcp_tag(self) -> str:
        """
        Represents the VLAN tag the DHCP Server is in/answering to.

        :getter: The VLAN tag or nothing if a system with the profile should not be in a VLAN.
        :setter: The new VLAN tag.
        """
        return self._dhcp_tag

    @dhcp_tag.setter
    def dhcp_tag(self, dhcp_tag: str):
        r"""
        Setter for the ``dhcp_tag`` property.

        :param dhcp_tag: The new VLAN tag.
        :raises TypeError: Raised in case the tag was not of type ``str``.
        """
        if not isinstance(dhcp_tag, str):
            raise TypeError("Field dhcp_tag of object profile needs to be of type str!")
        self._dhcp_tag = dhcp_tag

    @property
    def server(self) -> str:
        """
        Represents the hostname the Cobbler server is reachable by a client.

        :getter: The hostname of the Cobbler server.
        :setter: May raise a ``TypeError`` in case the new value is not a ``str``.
        """
        return self._server

    @server.setter
    def server(self, server: str):
        """
        Setter for the server property.

        :param server: If this is None or an emtpy string this will be reset to be inherited from the parent object.
        :raises TypeError: In case the new value was not of type ``str``.
        """
        if not isinstance(server, str):
            raise TypeError("Field server of object profile needs to be of type str!")
        if server == "":
            server = enums.VALUE_INHERITED
        self._server = server

    @property
    def next_server_v4(self) -> str:
        """
        Represents the next server for IPv4.

        :getter: The IP for the next server.
        :setter: May raise a ``TypeError`` if the new value is not of type ``str``.
        """
        return self._next_server_v4

    @next_server_v4.setter
    def next_server_v4(self, server: str = ""):
        """
        Setter for the next server value.

        :param server: The address of the IPv4 next server. Must be a string or ``enums.VALUE_INHERITED``.
        :raises TypeError: In case server is no string.
        """
        if not isinstance(server, str):
            raise TypeError("Server must be a string.")
        if server == enums.VALUE_INHERITED:
            self._next_server_v4 = enums.VALUE_INHERITED
        else:
            self._next_server_v4 = validate.ipv4_address(server)

    @property
    def next_server_v6(self) -> str:
        r"""
        Represents the next server for IPv6.

        :getter: The IP for the next server.
        :setter: May raise a ``TypeError`` if the new value is not of type ``str``.
        """
        return self._next_server_v6

    @next_server_v6.setter
    def next_server_v6(self, server: str = ""):
        """
        Setter for the next server value.

        :param server: The address of the IPv6 next server. Must be a string or ``enums.VALUE_INHERITED``.
        :raises TypeError: In case server is no string.
        """
        if not isinstance(server, str):
            raise TypeError("Server must be a string.")
        if server == enums.VALUE_INHERITED:
            self._next_server_v6 = enums.VALUE_INHERITED
        else:
            self._next_server_v6 = validate.ipv6_address(server)

    @property
    def filename(self) -> str:
        """
        The filename which is fetched by the client from TFTP.

        :getter: Either the default/inherited one, or the one specific to this profile.
        :setter: The new filename which is fetched on boot. May raise a ``TypeError`` when the wrong type was given.
        """
        return self._resolve("filename")

    @filename.setter
    def filename(self, filename: str):
        """
        The setter for the ``filename`` property.

        :param filename: The new ``filename`` for the profile.
        :raises TypeError: In case the new value was not of type ``str``.
        """
        if not isinstance(filename, str):
            raise TypeError("Field filename of object profile needs to be of type str!")
        if not filename:
            self._filename = enums.VALUE_INHERITED
        else:
            self._filename = filename.strip()

    @property
    def autoinstall(self) -> str:
        """
        Represents the automatic OS installation template file path, this must be a local file.

        :getter: Either the inherited name or the one specific to this profile.
        :setter: The name of the new autoinstall template is validated. The path should come in the format of a ``str``.
        """
        if self._autoinstall == enums.VALUE_INHERITED:
            parent = self.parent
            if parent is not None and isinstance(parent, Profile):
                return self.parent.autoinstall
            elif parent is not None and isinstance(parent, Distro):
                return self.api.settings().autoinstall
            else:
                self.logger.info("Profile \"%s\" did not have a valid parent of type Profile but autoinstall is set to "
                                 "\"<<inherit>>\".", self.name)
                return ""
        return self._autoinstall

    @autoinstall.setter
    def autoinstall(self, autoinstall: str):
        """
        Setter for the ``autoinstall`` property.

        :param autoinstall: local automatic installation template path
        """
        autoinstall_mgr = autoinstall_manager.AutoInstallationManager(self.api._collection_mgr)
        self._autoinstall = autoinstall_mgr.validate_autoinstall_template_file_path(autoinstall)

    @property
    def virt_auto_boot(self) -> bool:
        """
        Whether the VM should be booted when booting the host or not.

        :getter: ``True`` means autoboot is enabled, otherwise VM is not booted automatically.
        :setter: The new state for the property.
        """
        return self._virt_auto_boot

    @virt_auto_boot.setter
    def virt_auto_boot(self, num: bool):
        """
        Setter for booting a virtual machine automatically.

        :param num: The new value for whether to enable it or not.
        """
        self._virt_auto_boot = validate.validate_virt_auto_boot(num)

    @property
    def virt_cpus(self) -> int:
        """
        The amount of vCPU cores used in case the image is being deployed on top of a VM host.

        :getter: The cores used.
        :setter: The new number of cores.
        """
        return self._virt_cpus

    @virt_cpus.setter
    def virt_cpus(self, num: Union[int, str]):
        """
        Setter for the number of virtual CPU cores to assign to the virtual machine.

        :param num: The number of cpu cores.
        """
        self._virt_cpus = validate.validate_virt_cpus(num)

    @property
    def virt_file_size(self) -> int:
        r"""
        The size of the image and thus the usable size for the guest.

        .. warning:: There is a regression which makes the usage of multiple disks not possible right now. This will be
                     fixed in a future release.

        :getter: The size of the image(s) in GB.
        :setter: The float with the new size in GB.
        """
        return self._virt_file_size

    @virt_file_size.setter
    def virt_file_size(self, num: Union[str, int, float]):
        """
        Setter for the size of the virtual image size.

        :param num: The new size of the image.
        """
        self._virt_file_size = validate.validate_virt_file_size(num)

    @property
    def virt_disk_driver(self) -> enums.VirtDiskDrivers:
        """
        The type of disk driver used for storing the image.

        :getter: The enum type representation of the disk driver.
        :setter: May be a ``str`` with the name of the disk driver or from the enum type directly.
        """
        return self._virt_disk_driver

    @virt_disk_driver.setter
    def virt_disk_driver(self, driver: str):
        """
        Setter for the virtual disk driver that will be used.

        :param driver: The new driver.
        """
        self._virt_disk_driver = validate.validate_virt_disk_driver(driver)

    @property
    def virt_ram(self) -> int:
        """
        The amount of RAM given to the guest in MB.

        :getter: The amount of RAM currently assigned to the image.
        :setter: The new amount of ram. Must be an integer.
        """
        return self._virt_ram

    @virt_ram.setter
    def virt_ram(self, num: Union[str, int]):
        """
        Setter for the virtual RAM used for the VM.

        :param num: The number of RAM to use for the VM.
        """
        self._virt_ram = validate.validate_virt_ram(num)

    @property
    def virt_type(self) -> enums.VirtType:
        """
        The type of image used.

        :getter: The value of the virtual machine.
        :setter: May be of the enum type or a str which is then converted to the enum type.
        """
        return self._virt_type

    @virt_type.setter
    def virt_type(self, vtype: str):
        """
        Setter for the virtual machine type.

        :param vtype: May be on out of "qemu", "kvm", "xenpv", "xenfv", "vmware", "vmwarew", "openvz" or "auto".
        """
        self._virt_type = validate.validate_virt_type(vtype)

    @property
    def virt_bridge(self) -> str:
        """
        Represents the name of the virtual bridge to use.

        :getter: Either the default name for the bridge or the specific one for this profile.
        :setter: The new name. Does not overwrite the default one.
        """
        if not self._virt_bridge:
            return self.api.settings().default_virt_bridge
        return self._virt_bridge

    @virt_bridge.setter
    def virt_bridge(self, vbridge: str):
        """
        Setter for the name of the virtual bridge to use.

        :param vbridge: The name of the virtual bridge to use.
        """
        self._virt_bridge = validate.validate_virt_bridge(vbridge)

    @property
    def virt_path(self) -> str:
        """
        The path to the place where the image will be stored.

        :getter: The path to the image.
        :setter: The new path for the image.
        """
        return self._virt_path

    @virt_path.setter
    def virt_path(self, path: str):
        """
        Setter for the ``virt_path`` property.

        :param path: The path to where the image will be stored.
        """
        self._virt_path = validate.validate_virt_path(path)

    @property
    def repos(self) -> list:
        """
        The repositories to add once the system is provisioned.

        :getter: The names of the repositories the profile has assigned.
        :setter: The new names of the repositories for the profile. Validated against existing repositories.
        """
        return self._repos

    @repos.setter
    def repos(self, repos: list):
        """
        Setter of the repositories for the profile.

        :param repos: The new repositories which will be set.
        """
        self._repos = validate.validate_repos(repos, self.api, bypass_check=False)

    @property
    def redhat_management_key(self) -> str:
        """
        Getter of the redhat management key of the profile or it's parent.

        :getter: Returns the redhat_management_key of the profile.
        :setter: May raise a ``TypeError`` in case of a validation error.
        """
        return self._resolve("redhat_management_key")

    @redhat_management_key.setter
    def redhat_management_key(self, management_key: str):
        """
        Setter of the redhat management key.

        :param management_key: The value may be reset by setting it to None.
        """
        if not isinstance(management_key, str):
            raise TypeError("Field management_key of object profile is of type str!")
        if not management_key:
            self._redhat_management_key = enums.VALUE_INHERITED
        self._redhat_management_key = management_key

    @property
    def boot_loaders(self) -> list:
        """
        This represents all boot loaders for which Cobbler will try to generate bootloader configuration for.

        :getter: The bootloaders.
        :setter: The new bootloaders. Will be validates against a list of well known ones.
        """
        return self._resolve("boot_loaders")

    @boot_loaders.setter
    def boot_loaders(self, boot_loaders: list):
        """
        Setter of the boot loaders.

        :param boot_loaders: The boot loaders for the profile.
        :raises ValueError: In case the supplied boot loaders were not a subset of the valid ones.
        """
        if boot_loaders == enums.VALUE_INHERITED:
            self._boot_loaders = enums.VALUE_INHERITED
            return

        if boot_loaders:
            boot_loaders_split = utils.input_string_or_list(boot_loaders)

            parent = self.parent
            if parent is not None:
                parent_boot_loaders = parent.boot_loaders
            else:
                self.logger.warning("Parent of profile \"%s\" could not be found for resolving the parent bootloaders.",
                                    self.name)
                parent_boot_loaders = []
            if not set(boot_loaders_split).issubset(parent_boot_loaders):
                raise CX("Error with profile \"%s\" - not all boot_loaders are supported (given: \"%s\"; supported:"
                         "\"%s\")" % (self.name, str(boot_loaders_split), str(parent_boot_loaders)))
            self._boot_loaders = boot_loaders_split
        else:
            self._boot_loaders = []

    @property
    def menu(self) -> str:
        r"""
        Property to represent the menu which this image should be put into.

        :getter: The name of the menu or an emtpy str.
        :setter: Should only be the name of the menu not the object. May raise ``CX`` in case the menu does not exist.
        """
        return self._menu

    @menu.setter
    def menu(self, menu: str):
        """
        Setter for the menu property.

        :param menu: The menu for the image.
        :raises CX: In case the menu to be set could not be found.
        """
        if not isinstance(menu, str):
            raise TypeError("Field menu of object profile needs to be of type str!")
        if menu and menu != "":
            menu_list = self.api.menus()
            if not menu_list.find(name=menu):
                raise CX("menu %s not found" % menu)
        self._menu = menu

    @property
    def children(self) -> list:
        """
        This property represents all children of a distribution. It should not be set manually.

        :getter: The children of the distro.
        :setter: No validation is done because this is a Cobbler internal property.
        """
        return self._children

    @children.setter
    def children(self, value: list):
        """
        Setter for the children property.

        :param value: The new children of the distro.
        """
        self._children = value



"""
Cobbler daemon for logging remote syslog traffic during automatic installation

Copyright 2007-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import binascii
import logging.config
import os
import pwd
import time

from cobbler import remote, utils
from cobbler.api import CobblerAPI

if os.geteuid() == 0 and os.path.exists('/etc/cobbler/logging_config.conf'):
    logging.config.fileConfig('/etc/cobbler/logging_config.conf')


logger = logging.getLogger()


def core(cobbler_api: CobblerAPI):
    """
    Starts Cobbler.

    :param cobbler_api: The cobbler_api instance which is used for this method.
    """
    settings = cobbler_api.settings()
    xmlrpc_port = settings.xmlrpc_port

    regen_ss_file()
    do_xmlrpc_rw(cobbler_api, xmlrpc_port)


def regen_ss_file():
    """
    This is only used for Kerberos auth at the moment. It identifies XMLRPC requests from Apache that have already been
    cleared by Kerberos.
    """
    ssfile = "/var/lib/cobbler/web.ss"
    with open("/dev/urandom", 'rb') as fd:
        data = fd.read(512)

    with open(ssfile, 'wb', 0o660) as fd:
        fd.write(binascii.hexlify(data))

    http_user = "apache"
    family = utils.get_family()
    if family == "debian":
        http_user = "www-data"
    elif family == "suse":
        http_user = "wwwrun"
    os.lchown("/var/lib/cobbler/web.ss", pwd.getpwnam(http_user)[2], -1)


def do_xmlrpc_rw(cobbler_api: CobblerAPI, port):
    """
    This trys to bring up the Cobbler xmlrpc_api and restart it if it fails.

    :param cobbler_api: The cobbler_api instance which is used for this method.
    :param port: The port where the xmlrpc api should run on.
    """
    xinterface = remote.ProxiedXMLRPCInterface(cobbler_api, remote.CobblerXMLRPCInterface)
    server = remote.CobblerXMLRPCServer(('127.0.0.1', port))
    server.logRequests = 0      # don't print stuff
    logger.debug("XMLRPC running on %s", port)
    server.register_instance(xinterface)

    while True:
        try:
            logger.info("Cobbler startup complete")
            server.serve_forever()
        except IOError:
            # interrupted? try to serve again
            time.sleep(0.5)


if __name__ == "__main__":
    core(CobblerAPI())



"""
(C) 2008-2009, Red Hat Inc.
James Laska <jlaska@redhat.com>
Bill Peck <bpeck@redhat.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import glob
import os

from cobbler.cexceptions import CX


def register() -> str:
    """
    This pure python trigger acts as if it were a legacy shell-trigger, but is much faster. The return of this method
    indicates the trigger type.

    :return: Always ``/var/lib/cobbler/triggers/install/pre/*``
    """
    return "/var/lib/cobbler/triggers/install/pre/*"


def run(api, args) -> int:
    """
    The list of args should have one element:
        - 1: the name of the system or profile

    :param api: The api to resolve metadata with.
    :param args: This should be a list as described above.
    :return: "0" on success.
    :raises CX: Raised in case of missing arguments.
    """

    # FIXME: use the logger

    if len(args) < 3:
        raise CX("invalid invocation")

    name = args[1]

    settings = api.settings()

    # Remove any files matched with the given glob pattern
    def unlink_files(globex):
        for f in glob.glob(globex):
            if os.path.isfile(f):
                try:
                    os.unlink(f)
                except OSError:
                    pass

    if settings.anamon_enabled:
        dirname = "/var/log/cobbler/anamon/%s" % name
        if os.path.isdir(dirname):
            unlink_files(os.path.join(dirname, "*"))

    # TODO - log somewhere that we cleared a systems anamon logs
    return 0



"""
Cobbler's Mongo database based object serializer.
Experimental version.

Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>
James Cammarata <jimi@sngx.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

from configparser import ConfigParser

from cobbler import settings
from cobbler.cexceptions import CX

try:
    from pymongo import MongoClient
    from pymongo.errors import ConnectionFailure, ConfigurationError
    pymongo_loaded = True
except ModuleNotFoundError:
    # FIXME: log message
    pymongo_loaded = False

mongodb = None


def __connect(configfile: str = "/etc/cobbler/mongodb.conf"):
    """
    Reads the config file for mongodb and then connects to the mongodb.
    """
    cp = ConfigParser()
    cp.read(configfile)

    host = cp.get("connection", "host")
    port = int(cp.get("connection", "port"))
    # pylint: disable=global-statement
    global mongodb
    mongodb = MongoClient(host, port)['cobbler']
    try:
        # The ismaster command is cheap and doesn't require auth.
        mongodb.admin.command('ismaster')
    except ConnectionFailure as e:
        # FIXME: log error
        raise CX("Unable to connect to Mongo database or get database \"cobbler\"") from e
    except ConfigurationError as e:
        raise CX("The configuration of the MongoDB connection isn't correct, please check the Cobbler settings.") from e


def register() -> str:
    """
    The mandatory Cobbler module registration hook.
    """
    # FIXME: only run this if enabled.
    if not pymongo_loaded:
        return ""
    return "serializer"


def what() -> str:
    """
    Module identification function
    """
    return "serializer/mongodb"


def serialize_item(collection, item):
    """
    Save a collection item to database.

    :param collection: collection
    :param item: collection item
    """

    __connect()
    collection = mongodb[collection.collection_type()]
    data = collection.find_one({'name': item.name})
    if data:
        collection.update({'name': item.name}, item.serialize())
    else:
        collection.insert(item.serialize())


def serialize_delete(collection, item):
    """
    Delete a collection item from database.

    :param collection: collection
    :param item: collection item
    """

    __connect()
    collection = mongodb[collection.collection_type()]
    collection.remove({'name': item.name})


def serialize(collection):
    """
    Save a collection to database

    :param collection: collection
    """

    # TODO: error detection
    ctype = collection.collection_type()
    if ctype != "settings":
        for x in collection:
            serialize_item(collection, x)


def deserialize_raw(collection_type: str):
    """
    Get a collection from mongodb and parse it into an object.

    :param collection_type: The collection type to fetch.
    :return: The first element of the collection requested.
    """
    if collection_type == "settings":
        return settings.read_settings_file()
    else:
        __connect()
        collection = mongodb[collection_type]
        return collection.find()


def deserialize(collection, topological: bool = True):
    """
    Load a collection from the database.

    :param collection: The collection to deserialize.
    :param topological: If the collection list should be sorted by the collection dict depth value or not.
    """

    datastruct = deserialize_raw(collection.collection_type())
    if topological and type(datastruct) == list:
        datastruct.sort(key=lambda x: x["depth"])
    if type(datastruct) == dict:
        collection.from_dict(datastruct)
    elif type(datastruct) == list:
        collection.from_list(datastruct)



"""
Cobbler DownloadManager

Copyright 2018, Jorgen Maas <jorgen.maas@gmail.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import logging

import requests


class DownloadManager:

    def __init__(self, collection_mgr):
        """
        Constructor

        :param collection_mgr: This is the current collection manager instance which holds the settings.
        """
        self.collection_mgr = collection_mgr
        self.settings = collection_mgr.settings()
        self.logger = logging.getLogger()
        self.cert = ()
        if self.settings.proxy_url_ext:
            # requests wants a dict like:  protocol: proxy_uri
            self.proxies = self.settings.proxy_url_ext
        else:
            self.proxies = {}

    def urlread(self, url, proxies=None, cert=None):
        """
        Read the content of a given URL and pass the requests. Response object to the caller.

        :param url: The URL the request.
        :param proxies: Override the default Cobbler proxies.
        :param cert: Override the default Cobbler certs.
        :returns: The Python ``requests.Response`` object.
        """
        if proxies is None:
            proxies = self.proxies
        if cert is None:
            cert = self.cert
        return requests.get(url, proxies=proxies, cert=cert)

    def download_file(self, url, dst, proxies=None, cert=None):
        """
        Donwload a file from a URL and save it to any disc location.

        :param url: The URL the request.
        :param dst: The destination file path.
        :param proxies: Override the default Cobbler proxies.
        :param cert: Override the default Cobbler certs.
        """
        if proxies is None:
            proxies = self.proxies
        if cert is None:
            cert = self.cert
        response = requests.get(url, stream=True, proxies=proxies, cert=cert)
        with open(dst, "wb") as handle:
            for chunk in response.iter_content(chunk_size=512):
                # filter out keep-alive new chunks
                if chunk:
                    handle.write(chunk)



"""
Cobbler app-wide settings
"""
# SPDX-License-Identifier: GPL-2.0-or-later
# SPDX-FileCopyrightText: Copyright 2006-2008, Red Hat, Inc and Others
# SPDX-FileCopyrightText: Michael DeHaan <michael.dehaan AT gmail>
# SPDX-FileCopyrightText: 2021 Dominik Gedon <dgedon@suse.de>
# SPDX-FileCopyrightText: 2021 Enno Gotthold <egotthold@suse.de>
# SPDX-FileCopyrightText: Copyright SUSE LLC

import datetime
import glob
import logging
import os.path
import pathlib
import re
import shutil
import traceback
from pathlib import Path
from typing import Any, Dict, Hashable
import yaml
from schema import SchemaError, SchemaMissingKeyError, SchemaWrongKeyError

from cobbler import utils
from cobbler.settings import migrations

BIND_CHROOT_PATH = ""


class Settings:
    """
    This class contains all app-wide settings of Cobbler. It should only exist once in a Cobbler instance.
    """

    @staticmethod
    def collection_type() -> str:
        """
        This is a hardcoded string which represents the collection type.

        :return: "setting"
        """
        return "setting"

    @staticmethod
    def collection_types() -> str:
        """
        return the collection plural name
        """
        return "settings"

    def __init__(self):
        """
        Constructor.
        """
        self.allow_duplicate_hostnames = False
        self.allow_duplicate_ips = False
        self.allow_duplicate_macs = False
        self.allow_dynamic_settings = False
        self.always_write_dhcp_entries = False
        self.anamon_enabled = False
        self.auth_token_expiration = 3600
        self.authn_pam_service = "login"
        self.autoinstall_snippets_dir = "/var/lib/cobbler/snippets"
        self.autoinstall_templates_dir = "/var/lib/cobbler/templates"
        self.bind_chroot_path = BIND_CHROOT_PATH
        self.bind_zonefile_path = "/var/lib/named"
        self.bind_master = "127.0.0.1"
        self.boot_loader_conf_template_dir = "/etc/cobbler/boot_loader_conf"
        self.bootloaders_dir = "/var/lib/cobbler/loaders"
        self.grubconfig_dir = "/var/lib/cobbler/grub_config"
        self.build_reporting_enabled = False
        self.build_reporting_email = []
        self.build_reporting_ignorelist = []
        self.build_reporting_sender = ""
        self.build_reporting_smtp_server = "localhost"
        self.build_reporting_subject = ""
        self.buildisodir = "/var/cache/cobbler/buildiso"
        self.cheetah_import_whitelist = ["re", "random", "time"]
        self.client_use_https = False
        self.client_use_localhost = False
        self.cobbler_master = ""
        self.convert_server_to_ip = False
        self.createrepo_flags = "-c cache -s sha"
        self.autoinstall = "default.ks"
        self.default_name_servers = []
        self.default_name_servers_search = []
        self.default_ownership = ["admin"]
        self.default_password_crypted = r"\$1\$mF86/UHC\$WvcIcX2t6crBz2onWxyac."
        self.default_template_type = "cheetah"
        self.default_virt_bridge = "xenbr0"
        self.default_virt_disk_driver = "raw"
        self.default_virt_file_size = 5
        self.default_virt_ram = 512
        self.default_virt_type = "auto"
        self.enable_ipxe = False
        self.enable_menu = True
        self.http_port = 80
        self.include = ["/etc/cobbler/settings.d/*.settings"]
        self.iso_template_dir = "/etc/cobbler/iso"
        self.jinja2_includedir = "/var/lib/cobbler/jinja2"
        self.kernel_options = {}
        self.ldap_anonymous_bind = True
        self.ldap_base_dn = "DC=devel,DC=redhat,DC=com"
        self.ldap_port = 389
        self.ldap_search_bind_dn = ""
        self.ldap_search_passwd = ""
        self.ldap_search_prefix = 'uid='
        self.ldap_server = "grimlock.devel.redhat.com"
        self.ldap_tls = True
        self.ldap_tls_cacertfile = ""
        self.ldap_tls_certfile = ""
        self.ldap_tls_keyfile = ""
        self.bind_manage_ipmi = False
        # TODO: Remove following line
        self.manage_dhcp = False
        self.manage_dhcp_v6 = False
        self.manage_dhcp_v4 = False
        self.manage_dns = False
        self.manage_forward_zones = []
        self.manage_reverse_zones = []
        self.manage_genders = False
        self.manage_rsync = False
        self.manage_tftpd = True
        self.mgmt_classes = []
        self.mgmt_parameters = {}
        self.next_server_v4 = "127.0.0.1"
        self.next_server_v6 = "::1"
        self.nsupdate_enabled = False
        self.nsupdate_log = "/var/log/cobbler/nsupdate.log"
        self.nsupdate_tsig_algorithm = "hmac-sha512"
        self.nsupdate_tsig_key = []
        self.power_management_default_type = "ipmilanplus"
        self.proxy_url_ext = ""
        self.proxy_url_int = ""
        self.puppet_auto_setup = False
        self.puppet_parameterized_classes = True
        self.puppet_server = "puppet"
        self.puppet_version = 2
        self.puppetca_path = "/usr/bin/puppet"
        self.pxe_just_once = True
        self.nopxe_with_triggers = True
        self.redhat_management_permissive = False
        self.redhat_management_server = "xmlrpc.rhn.redhat.com"
        self.redhat_management_key = ""
        self.register_new_installs = False
        self.remove_old_puppet_certs_automatically = False
        self.replicate_repo_rsync_options = "-avzH"
        self.replicate_rsync_options = "-avzH"
        self.reposync_flags = "-l -m -d"
        self.reposync_rsync_flags = ""
        self.restart_dhcp = True
        self.restart_dns = True
        self.run_install_triggers = True
        self.scm_track_enabled = False
        self.scm_track_mode = "git"
        self.scm_track_author = "cobbler <cobbler@localhost>"
        self.scm_push_script = "/bin/true"
        self.serializer_pretty_json = False
        self.server = "127.0.0.1"
        self.sign_puppet_certs_automatically = False
        self.signature_path = "/var/lib/cobbler/distro_signatures.json"
        self.signature_url = "https://cobbler.github.io/signatures/3.0.x/latest.json"
        self.tftpboot_location = "/var/lib/tftpboot"
        self.virt_auto_boot = False
        self.webdir = "/var/www/cobbler"
        self.webdir_whitelist = [".link_cache", "misc", "distro_mirror", "images", "links", "localmirror", "pub",
                                 "rendered", "repo_mirror", "repo_profile", "repo_system", "svc", "web", "webui"]
        self.xmlrpc_port = 25151
        self.yum_distro_priority = 1
        self.yum_post_install_mirror = True
        self.yumdownloader_flags = "--resolve"
        self.windows_enabled = False
        self.windows_template_dir = "/etc/cobbler/windows"
        self.samba_distro_share = "DISTRO"

    def to_string(self) -> str:
        """
        Returns the kernel options as a string.

        :return: The multiline string with the kernel options.
        """
        buf = "defaults\n"
        buf += "kernel options  : %s\n" % self.__dict__['kernel_options']
        return buf

    def to_dict(self) -> dict:
        """
        Return an easily serializable representation of the config.

        .. deprecated:: 3.2.1
           Use ``obj.__dict__`` directly please. Will be removed with 3.3.0

        :return: The dict with all user settings combined with settings which are left to the default.
        """
        # TODO: Deprecate and remove. Tailcall is not needed.
        return self.__dict__

    def from_dict(self, new_values):
        """
        Modify this object to load values in dictionary. If the handed dict would lead to an invalid object it is
        silently discarded.

        .. warning:: If the dict from the args has not all settings included Cobbler may behave unexpectedly.

        :param new_values: The dictionary with settings to replace.
        :return: Returns the settings instance this method was called from.
        """
        if new_values is None:
            logging.warning("Not loading empty settings dictionary!")
            return

        old_settings = self.__dict__

        self.__dict__.update(new_values)

        if not self.is_valid():
            self.__dict__ = old_settings

        return self

    def is_valid(self) -> bool:
        """
        Silently drops all errors and returns ``True`` when everything is valid.

        :return: If this settings object is valid this returns true. Otherwise false.
        """
        try:
            validate_settings(self.__dict__)
        except SchemaError:
            return False
        return True

    def __getattr__(self, name):
        """
        This returns the current value of the setting named in the args.

        :param name: The setting to return the value of.
        :return: The value of the setting "name".
        """
        try:
            if name == "kernel_options":
                # backwards compatibility -- convert possible string value to dict
                (success, result) = utils.input_string_or_dict(self.__dict__[name], allow_multiples=False)
                self.__dict__[name] = result
                return result
            # TODO: This needs to be explicitly tested
            elif name == "manage_dhcp":
                return self.manage_dhcp_v4
            return self.__dict__[name]
        except Exception as error:
            if name in self.__dict__:
                return self.__dict__[name]
            else:
                raise AttributeError(f"no settings attribute named '{name}' found") from error

    def save(self, filepath="/etc/cobbler/settings.yaml"):
        """
        Saves the settings to the disk.
        """
        update_settings_file(self.to_dict(), filepath)


def validate_settings(settings_content: dict) -> dict:
    """
    This function performs logical validation of our loaded YAML files.
    This function will:
    - Perform type validation on all values of all keys.
    - Provide defaults for optional settings.
    :param settings_content: The dictionary content from the YAML file.
    :raises SchemaError: In case the data given is invalid.
    :return: The Settings of Cobbler which can be safely used inside this instance.
    """
    return migrations.normalize(settings_content)


def parse_bind_config(configpath: str):
    """
    Parse the Bind9 configuration file and adjust the Cobbler default settings according to the readings.

    :param configpath: The path in the filesystem where the file can be read.
    """
    global BIND_CHROOT_PATH
    bind_config = {}
    # When running as a webapp we can't access this, but don't need it
    try:
        bind_config_file = open(configpath, "r")
    except (IOError, OSError):
        pass
    else:
        for line in bind_config_file:
            if re.match(r"[a-zA-Z]+=", line):
                (name, value) = line.rstrip().split("=")
                bind_config[name] = value.strip('"')
        # RHEL, SysV Fedora
        if "ROOTDIR" in bind_config:
            BIND_CHROOT_PATH = bind_config["ROOTDIR"]
        # Debian, Systemd Fedora
        if "OPTIONS" in bind_config:
            rootdirmatch = re.search(r"-t ([/\w]+)", bind_config["OPTIONS"])
            if rootdirmatch is not None:
                BIND_CHROOT_PATH = rootdirmatch.group(1)


def autodetect_bind_chroot():
    """
    Autodetect bind chroot configuration
    """
    bind_config_filename = None
    if os.path.exists("/etc/sysconfig/named"):
        # RHEL/Fedora
        bind_config_filename = "/etc/sysconfig/named"
    else:
        # Debian
        bind_config_files = glob.glob("/etc/default/bind*")
        for filename in bind_config_files:
            if os.path.exists(filename):
                bind_config_filename = filename
    # Parse the config file if available
    if bind_config_filename:
        parse_bind_config(bind_config_filename)


def read_yaml_file(filepath ="/ect/cobbler/settings.yaml") -> Dict[Hashable, Any]:
    """
    Reads settings files from ``filepath`` and all paths in `include` (which is read from the settings file) and saves
    the content in a dictionary.
    Any key may be overwritten in a later loaded settings file. The last loaded file wins.

    :param filepath: Settings file path, defaults to "/ect/cobbler/settings.yaml"
    :raises FileNotFoundError: In case file does not exist or is a directory.
    :raises yaml.YAMLError: In case the file is not a valid YAML file.
    :return: The aggregated dict of all settings.
    """
    if not os.path.isfile(filepath):
        raise FileNotFoundError('Given path "%s" does not exist or is a directory.' % filepath)
    try:
        with open(filepath) as main_settingsfile:
            filecontent = yaml.safe_load(main_settingsfile.read())

            for ival in filecontent.get("include", []):
                for ifile in glob.glob(ival):
                    with open(ifile, 'r') as extra_settingsfile:
                        filecontent.update(yaml.safe_load(extra_settingsfile.read()))
    except yaml.YAMLError as error:
        traceback.print_exc()
        raise yaml.YAMLError('"%s" is not a valid YAML file' % filepath) from error
    return filecontent


def read_settings_file(filepath="/etc/cobbler/settings.yaml") -> Dict[Hashable, Any]:
    """
    Utilizes ``read_yaml_file()``. If the read settings file is invalid in the context of Cobbler we will return an
    empty dictionary.

    :param filepath: The path to the settings file.
    :raises SchemaMissingKeyError: In case keys are minssing.
    :raises SchemaWrongKeyError: In case keys are not listed in the schema.
    :raises SchemaError: In case the schema is wrong.
    :return: A dictionary with the settings. As a word of caution: This may not represent a correct settings object, it
             will only contain a correct YAML representation.
    """
    filecontent = read_yaml_file(filepath)

    # FIXME: Do not call validate_settings() because of chicken - egg problem
    try:
        validate_settings(filecontent)
    except SchemaMissingKeyError:
        logging.exception("Settings file was not returned due to missing keys.")
        logging.debug('The settings to read were: "%s"', filecontent)
        return {}
    except SchemaWrongKeyError:
        logging.exception("Settings file was returned due to an error in the schema.")
        logging.debug('The settings to read were: "%s"', filecontent)
        return {}
    except SchemaError:
        logging.exception("Settings file was returned due to an error in the schema.")
        logging.debug('The settings to read were: "%s"', filecontent)
        return {}
    return filecontent


def update_settings_file(data: dict, filepath="/etc/cobbler/settings.yaml") -> bool:
    """
    Write data handed to this function into the settings file of Cobbler. This function overwrites the existing content.
    It will only write valid settings. If you are trying to save invalid data this will raise a SchemaException
    described in :py:meth:`cobbler.settings.validate`.

    :param data: The data to put into the settings file.
    :param filepath: This sets the path of the settingsfile to write.
    :return: True if the action succeeded. Otherwise return False.
    """
    # Backup old settings file
    path = pathlib.Path(filepath)
    if path.exists():
        timestamp = str(datetime.datetime.now().strftime("%Y%m%d_%H-%M-%S"))
        shutil.copy(path, path.parent.joinpath(f"{path.stem}_{timestamp}{path.suffix}"))

    try:
        validated_data = validate_settings(data)
        with open(filepath, "w") as settings_file:
            yaml_dump = yaml.safe_dump(validated_data)
            header = "# Cobbler settings file\n"
            header += "# Docs for this file can be found at: https://cobbler.readthedocs.io/en/latest/cobbler-conf.html"
            header += "\n\n"
            yaml_dump = header + yaml_dump
            settings_file.write(yaml_dump)
        return True
    except SchemaMissingKeyError:
        logging.exception("Settings file was not written to the disc due to missing keys.")
        logging.debug('The settings to write were: "%s"', data)
        return False
    except SchemaError:
        logging.exception("Settings file was not written to the disc due to an error in the schema.")
        logging.debug('The settings to write were: "%s"', data)
        return False


def migrate(yaml_dict: dict, settings_path: Path) -> dict:
    """
    Migrates the current settings

    :param yaml_dict: The settings dict
    :param settings_path: The settings path
    :return: The migrated settings
    """
    return migrations.migrate(yaml_dict, settings_path)


# Initialize Settings module for manipulating the global DEFAULTS variable
autodetect_bind_chroot()



"""
This module contains code in order to create the automatic installation files. For example kickstarts, autoyast files
or preseed files.
"""

import logging
import os

from cobbler import autoinstallgen
from cobbler import utils

TEMPLATING_ERROR = 1
KICKSTART_ERROR = 2


class AutoInstallationManager:
    """
    Manage automatic installation templates, snippets and final files
    """

    def __init__(self, collection_mgr):
        """
        Constructor for the autoinstall manager.

        :param collection_mgr: The collection manager which has all objects.
        """

        self.collection_mgr = collection_mgr
        self.snippets_base_dir = self.collection_mgr.settings().autoinstall_snippets_dir
        self.templates_base_dir = self.collection_mgr.settings().autoinstall_templates_dir
        self.autoinstallgen = autoinstallgen.AutoInstallationGen(self.collection_mgr)
        self.logger = logging.getLogger()

    def validate_autoinstall_template_file_path(self, autoinstall: str, for_item: bool = True,
                                                new_autoinstall: bool = False) -> str:
        """
        Validate the automatic installation template's relative file path.

        :param autoinstall: automatic installation template relative file path
        :param for_item: enable/disable special handling for Item objects
        :param new_autoinstall: when set to true new filenames are allowed
        :returns: automatic installation template relative file path
        :raises TypeError: Raised in case ``autoinstall`` is not a string.
        :raises OSError: Raised in case template file not found.
        :raises ValueError: Raised in case template file is invalid.
        """

        if not isinstance(autoinstall, str):
            raise TypeError("Invalid input, autoinstall must be a string")
        autoinstall = autoinstall.strip()

        if autoinstall == "":
            # empty autoinstall is allowed (interactive installations)
            return autoinstall

        if for_item is True:
            # this autoinstall value has special meaning for Items
            # other callers of this function have no use for this
            if autoinstall == "<<inherit>>":
                return autoinstall

        if autoinstall.find("..") != -1:
            raise ValueError("Invalid automatic installation template file location %s, it must not contain .."
                             % autoinstall)

        autoinstall_path = "%s/%s" % (self.templates_base_dir, autoinstall)
        if not os.path.isfile(autoinstall_path) and not new_autoinstall:
            raise OSError("Invalid automatic installation template file location %s, file not found" % autoinstall_path)

        return autoinstall

    def get_autoinstall_templates(self) -> list:
        """
        Get automatic OS installation templates

        :returns: A list of automatic installation templates
        """

        files = []
        for root, dirnames, filenames in os.walk(self.templates_base_dir):
            for filename in filenames:
                rel_root = root[len(self.templates_base_dir) + 1:]
                if rel_root:
                    rel_path = "%s/%s" % (rel_root, filename)
                else:
                    rel_path = filename
                files.append(rel_path)

        files.sort()
        return files

    def read_autoinstall_template(self, file_path: str) -> str:
        """
        Read an automatic OS installation template

        :param file_path: automatic installation template relative file path
        :returns: automatic installation template content
        """

        file_path = self.validate_autoinstall_template_file_path(file_path, for_item=False)

        file_full_path = "%s/%s" % (self.templates_base_dir, file_path)
        fileh = open(file_full_path, "r")
        data = fileh.read()
        fileh.close()

        return data

    def write_autoinstall_template(self, file_path: str, data: str) -> bool:
        """
        Write an automatic OS installation template

        :param file_path: automatic installation template relative file path
        :param data: automatic installation template content
        """

        file_path = self.validate_autoinstall_template_file_path(file_path, for_item=False, new_autoinstall=True)

        file_full_path = "%s/%s" % (self.templates_base_dir, file_path)
        try:
            utils.mkdir(os.path.dirname(file_full_path))
        except:
            utils.die("unable to create directory for automatic OS installation template at %s" % file_path)

        fileh = open(file_full_path, "w+")
        fileh.write(data)
        fileh.close()

        return True

    def remove_autoinstall_template(self, file_path: str):
        """
        Remove an automatic OS installation template

        :param file_path: automatic installation template relative file path
        """

        file_path = self.validate_autoinstall_template_file_path(file_path, for_item=False)

        file_full_path = "%s/%s" % (self.templates_base_dir, file_path)
        if not self.is_autoinstall_in_use(file_path):
            os.remove(file_full_path)
        else:
            utils.die("attempt to delete in-use file")

    def validate_autoinstall_snippet_file_path(self, snippet: str, new_snippet: bool = False) -> str:
        """
        Validate the snippet's relative file path.

        :param snippet: automatic installation snippet relative file path
        :param new_snippet: when set to true new filenames are allowed
        :returns: Snippet if successful otherwise raises an exception.
        :raises TypeError: Raised in case ``snippet`` is not a string.
        :raises ValueError: Raised in case snippet file is invalid.
        :raises OSError: Raised in case snippet file location is not found.
        """

        if not isinstance(snippet, str):
            raise TypeError("Invalid input, snippet must be a string")
        snippet = snippet.strip()

        if snippet.find("..") != -1:
            raise ValueError("Invalid automated installation snippet file location %s, it must not contain .."
                             % snippet)

        snippet_path = "%s/%s" % (self.snippets_base_dir, snippet)
        if not os.path.isfile(snippet_path) and not new_snippet:
            raise OSError("Invalid automated installation snippet file location %s, file not found" % snippet_path)

        return snippet

    def get_autoinstall_snippets(self) -> list:
        """
        Get a list of all autoinstallation snippets.

        :return: The list of snippets
        """
        files = []
        for root, dirnames, filenames in os.walk(self.snippets_base_dir):

            for filename in filenames:
                rel_root = root[len(self.snippets_base_dir) + 1:]
                if rel_root:
                    rel_path = "%s/%s" % (rel_root, filename)
                else:
                    rel_path = filename
                files.append(rel_path)

        files.sort()
        return files

    def read_autoinstall_snippet(self, file_path: str) -> str:
        """
        Reads a autoinstall snippet from underneath the configured snippet base dir.

        :param file_path: The relative file path under the configured snippets base dir.
        :return: The read snippet.
        """
        file_path = self.validate_autoinstall_snippet_file_path(file_path)

        file_full_path = "%s/%s" % (self.snippets_base_dir, file_path)
        fileh = open(file_full_path, "r")
        data = fileh.read()
        fileh.close()

        return data

    def write_autoinstall_snippet(self, file_path: str, data: str):
        """
        Writes a snippet with the given content to the relative path under the snippet root directory.

        :param file_path: The relative path under the configured snippet base dir.
        :param data: The snippet code.
        """
        file_path = self.validate_autoinstall_snippet_file_path(file_path, new_snippet=True)

        file_full_path = "%s/%s" % (self.snippets_base_dir, file_path)
        try:
            utils.mkdir(os.path.dirname(file_full_path))
        except:
            utils.die("unable to create directory for automatic OS installation snippet at %s" % file_path)

        fileh = open(file_full_path, "w+")
        fileh.write(data)
        fileh.close()

    def remove_autoinstall_snippet(self, file_path: str) -> bool:
        """
        Remove the autoinstall snippet with the given path.

        :param file_path: The path relative to the configured snippet root.
        :return: A boolean indicating the success of the task.
        """
        file_path = self.validate_autoinstall_snippet_file_path(file_path)

        file_full_path = "%s/%s" % (self.snippets_base_dir, file_path)
        os.remove(file_full_path)

        return True

    def is_autoinstall_in_use(self, name: str) -> bool:
        """
        Reports the status if a given system is currently being provisioned.

        :param name: The name of the system.
        :return: Whether the system is in install mode or not.
        """
        for x in self.collection_mgr.profiles():
            if x.autoinstall is not None and x.autoinstall == name:
                return True
        for x in self.collection_mgr.systems():
            if x.autoinstall is not None and x.autoinstall == name:
                return True
        return False

    def generate_autoinstall(self, profile=None, system=None) -> str:
        """
        Generates the autoinstallation for a system or a profile. You may only specifify one parameter. If you specify
        both, the system is generated and the profile argument is ignored.

        :param profile: The Cobbler profile you want an autoinstallation generated for.
        :param system: The Cobbler system you want an autoinstallation generated for.
        :return: The rendered template for the system or profile.
        """
        if system is not None:
            return self.autoinstallgen.generate_autoinstall_for_system(system)
        elif profile is not None:
            return self.autoinstallgen.generate_autoinstall_for_profile(profile)

    def log_autoinstall_validation_errors(self, errors_type: int, errors: list):
        """
        Log automatic installation file errors

        :param errors_type: validation errors type
        :param errors: A list with all the errors which occurred.
        """

        if errors_type == TEMPLATING_ERROR:
            self.logger.warning("Potential templating errors:")
            for error in errors:
                (line, col) = error["lineCol"]
                line -= 1   # we add some lines to the template data, so numbering is off
                self.logger.warning("Unknown variable found at line %d, column %d: '%s'", line, col, error["rawCode"])
        elif errors_type == KICKSTART_ERROR:
            self.logger.warning("Kickstart validation errors: %s", errors[0])

    def validate_autoinstall_file(self, obj, is_profile: bool) -> list:
        """
        Validate automatic installation file used by a system/profile.

        :param obj: system/profile
        :param is_profile: if obj is a profile
        :returns: [bool, int, list] list with validation result, errors type and list of errors
        """

        blended = utils.blender(self.collection_mgr.api, False, obj)

        # get automatic installation template
        autoinstall = blended["autoinstall"]
        if autoinstall is None or autoinstall == "":
            self.logger.info("%s has no automatic installation template set, skipping", obj.name)
            return [True, 0, ()]

        # generate automatic installation file
        os_version = blended["os_version"]
        self.logger.info("----------------------------")
        self.logger.debug("osversion: %s", os_version)
        if is_profile:
            self.generate_autoinstall(profile=obj)
        else:
            self.generate_autoinstall(system=obj)
        last_errors = self.autoinstallgen.get_last_errors()
        if len(last_errors) > 0:
            return [False, TEMPLATING_ERROR, last_errors]
        return [True, 0, ()]

    def validate_autoinstall_files(self) -> bool:
        """
        Determine if Cobbler automatic OS installation files will be accepted by corresponding Linux distribution
        installers. The presence of an error does not imply that the automatic installation file is bad, only that the
        possibility exists. Automatic installation file validators are not available for all automatic installation file
        types and on all operating systems in which Cobbler may be installed.

        :return: True if all automatic installation files are valid, otherwise false.
        """
        overall_success = True

        for x in self.collection_mgr.profiles():
            (success, errors_type, errors) = self.validate_autoinstall_file(x, True)
            if not success:
                overall_success = False
            if len(errors) > 0:
                self.log_autoinstall_validation_errors(errors_type, errors)
        for x in self.collection_mgr.systems():
            (success, errors_type, errors) = self.validate_autoinstall_file(x, False)
            if not success:
                overall_success = False
            if len(errors) > 0:
                self.log_autoinstall_validation_errors(errors_type, errors)

        if not overall_success:
            self.logger.warning("*** potential errors detected in automatic installation files ***")
        else:
            self.logger.info("*** all automatic installation files seem to be ok ***")

        return overall_success



"""
Module loader, adapted for Cobbler usage

Copyright 2006-2009, Red Hat, Inc and Others
Adrian Likins <alikins@redhat.com>
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import logging
from configparser import ConfigParser
from importlib import import_module

import glob
import os
from typing import Optional, Dict, Any

from cobbler.cexceptions import CX
from cobbler.utils import log_exc

# add cobbler/modules to python path
import cobbler

# TODO: add os.path.normpath()
mod_path = os.path.join(os.path.abspath(os.path.dirname(cobbler.__file__)), 'modules')

MODULE_CACHE: Dict[str, Any] = {}
MODULES_BY_CATEGORY: Dict[str, Dict[str, Any]] = {}


logger = logging.getLogger()


def load_modules(module_path: str = mod_path):
    """
    Load the modules from the path handed to the function into Cobbler.

    :param module_path: The path which should be considered as the root module path.
    :return: Two dictionary's with the dynamically loaded modules.
    """

    filenames = glob.glob("%s/*.py" % module_path)
    filenames += glob.glob("%s/*.pyc" % module_path)
    filenames += glob.glob("%s/*.pyo" % module_path)
    # Allow recursive modules
    filenames += glob.glob("%s/**/*.py" % module_path)
    filenames += glob.glob("%s/**/*.pyc" % module_path)
    filenames += glob.glob("%s/**/*.pyo" % module_path)

    for fn in filenames:
        # FIXME: Use module_path instead of mod_path
        basename = fn.replace(mod_path, '')
        modname = ""

        if basename.__contains__("__pycache__") or basename.__contains__("__init__.py"):
            continue

        if basename[0] == "/":
            basename = basename[1:]

        basename = basename.replace("/", ".")

        if basename[-3:] == ".py":
            modname = basename[:-3]
        elif basename[-4:] in [".pyc", ".pyo"]:
            modname = basename[:-4]

        # FIXME: Use module_path instead of mod_path
        __import_module(mod_path, modname)

    return MODULE_CACHE, MODULES_BY_CATEGORY


def __import_module(module_path: str, modname: str):
    """
    Import a module which is not part of the core functionality of Cobbler.

    :param module_path: The path to the module.
    :param modname: The name of the module.
    """
    try:
        blip = import_module("cobbler.modules.%s" % modname)
        if not hasattr(blip, "register"):
            if not modname.startswith("__init__"):
                errmsg = "%(module_path)s/%(modname)s is not a proper module"
                print(errmsg % {'module_path': module_path, 'modname': modname})
            return None
        category = blip.register()
        if category:
            MODULE_CACHE[modname] = blip
        if category not in MODULES_BY_CATEGORY:
            MODULES_BY_CATEGORY[category] = {}
        MODULES_BY_CATEGORY[category][modname] = blip
    except Exception:
        logger.info('Exception raised when loading module %s' % modname)
        log_exc()


def get_module_by_name(name: str):
    """
    Get a module by its name. The category of the module is not needed.

    :param name: The name of the module.
    :return: The module asked by the function parameter.
    """
    return MODULE_CACHE.get(name, None)


def get_module_name(category: str, field: str, fallback_module_name: Optional[str] = None) -> str:
    """
    Get module name from configuration file (currently hardcoded ``/etc/cobbler/modules.conf``).

    :param category: Field category in configuration file.
    :param field: Field in configuration file
    :param fallback_module_name: Default value used if category/field is not found in configuration file
    :raises FileNotFoundError: If unable to find configuration file.
    :raises ValueError: If the category does not exist or the field is empty.
    :raises CX: If the field could not be read and no fallback_module_name was given.
    :returns: The name of the module.
    """
    modules_conf_path = "/etc/cobbler/modules.conf"
    if not os.path.exists(modules_conf_path):
        raise FileNotFoundError("Configuration file at \"%s\" not found" % modules_conf_path)

    cp = ConfigParser()
    cp.read(modules_conf_path)

    # FIXME: We can't enabled this check since it is to strict atm.
    # if category not in MODULES_BY_CATEGORY:
    # raise ValueError("category must be one of: %s" % MODULES_BY_CATEGORY.keys())

    if field.isspace():
        raise ValueError("field cannot be empty. Did you mean \"module\" maybe?")

    try:
        value = cp.get(category, field)
    except:
        if fallback_module_name is not None:
            value = fallback_module_name
        else:
            raise CX("Cannot find config file setting for: %s" % field)
    return value


def get_module_from_file(category: str, field: str, fallback_module_name: Optional[str] = None):
    """
    Get Python module, based on name defined in configuration file

    :param category: field category in configuration file
    :param field: field in configuration file
    :param fallback_module_name: default value used if category/field is not found in configuration file
    :raises CX: If unable to load Python module
    :returns: A Python module.
    """

    module_name = get_module_name(category, field, fallback_module_name)
    requested_module = MODULE_CACHE.get(module_name, None)
    if requested_module is None:
        raise CX("Failed to load module for %s/%s" % (category, field))
    return requested_module


def get_modules_in_category(category: str) -> list:
    """
    Return all modules of a module category.

    :param category: The module category.
    :return: A list of all modules of that category. Returns an empty list if the Category does not exist.
    """
    if category not in MODULES_BY_CATEGORY:
        # FIXME: We can't enabled this check since it is to strict atm.
        # raise ValueError("category must be one of: %s" % MODULES_BY_CATEGORY.keys())
        return []
    return list(MODULES_BY_CATEGORY[category].values())



"""
Authorization module that allow users listed in
/etc/cobbler/users.conf to be permitted to access resources, with
the further restriction that Cobbler objects can be edited to only
allow certain users/groups to access those specific objects.

Copyright 2008-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""


from configparser import ConfigParser

import os
from typing import Dict, Union

from cobbler.cexceptions import CX


def register() -> str:
    """
    The mandatory Cobbler module registration hook.

    :return: Always "authz"
    """
    return "authz"


def __parse_config() -> Dict[str, dict]:
    """
    Parse the "users.conf" of Cobbler and return all data in a dictionary.

    :return: The data separated by sections. Each section has a subdictionary with the key-value pairs.
    :raises FileNotFoundError
    """
    etcfile = '/etc/cobbler/users.conf'
    if not os.path.exists(etcfile):
        raise FileNotFoundError("/etc/cobbler/users.conf does not exist")
    # Make users case sensitive to handle kerberos
    config = ConfigParser()
    config.optionxform = str
    config.read(etcfile)
    alldata = {}
    sections = config.sections()
    for g in sections:
        alldata[str(g)] = {}
        opts = config.options(g)
        for o in opts:
            alldata[g][o] = 1
    return alldata


def __authorize_autoinst(api_handle, groups, user, autoinst) -> int:
    """
    The authorization rules for automatic installation file editing are a bit of a special case. Non-admin users can
    edit a automatic installation file only if all objects that depend on that automatic installation file are editable
    by the user in question.

    Example:
      if Pinky owns ProfileA
      and the Brain owns ProfileB
      and both profiles use the same automatic installation template
      and neither Pinky nor the Brain is an admin
      neither is allowed to edit the automatic installation template
      because they would make unwanted changes to each other

    In the above scenario the UI will explain the problem and ask that the user asks the admin to resolve it if
    required.
    NOTE: this function is only called by authorize so admin users are cleared before this function is called.

    :param api_handle: The api to resolve required information.
    :param groups: The groups a user is in.
    :param user: The user which is asking for access.
    :param autoinst: The automatic installation in question.
    :return: ``1`` if the user is allowed and otherwise ``0``.
    """

    lst = api_handle.find_profile(autoinst=autoinst, return_list=True)
    lst.extend(api_handle.find_system(autoinst=autoinst, return_list=True))
    for obj in lst:
        if not __is_user_allowed(obj, groups, user, "write_autoinst", autoinst, None):
            return 0
    return 1


def __authorize_snippet(api_handle, groups, user, autoinst) -> bool:
    """
    Only allow admins to edit snippets -- since we don't have detection to see where each snippet is in use.

    :param api_handle: Unused parameter.
    :param groups: The group which is asking for access.
    :param user: Unused parameter.
    :param autoinst: Unused parameter.
    :return: ``True`` if the group is allowed, otherwise ``False``.
    """

    for group in groups:
        if group not in ["admins", "admin"]:
            return False
    return True


def __is_user_allowed(obj, groups, user, resource, arg1, arg2) -> Union[int, bool]:
    """
    Check if a user is allowed to access the resource in question.

    :param obj: The object which is in question.
    :param groups: The groups a user is belonging to.
    :param user: The user which is demanding access to the ``obj``.
    :param resource: Unused parameter.
    :param arg1: Unused parameter.
    :param arg2: Unused parameter.
    :return: ``True`` if user is allowed, otherwise ``0``.
    """

    if user == "<DIRECT>":
        # system user, logged in via web.ss
        return True
    for group in groups:
        if group in ["admins", "admin"]:
            return True
    if obj.owners == []:
        return True
    for allowed in obj.owners:
        if user == allowed:
            # user match
            return True
        # else look for a group match
    for group in groups:
        if group == allowed:
            return True
    return 0


def authorize(api_handle, user: str, resource: str, arg1=None, arg2=None) -> Union[bool, int]:
    """
    Validate a user against a resource. All users in the file are permitted by this module.

    :param api_handle: The api to resolve required information.
    :param user: The user to authorize to the resource.
    :param resource: The resource the user is asking for access. This is something abstract like a remove operation.
    :param arg1: This is normally the name of the specific object in question.
    :param arg2: This parameter is pointless currently. Reserved for future code.
    :return: ``True`` or ``1`` if okay, otherwise ``False``.
    """
    if user == "<DIRECT>":
        # CLI should always be permitted
        return True

    # Everybody can get read-only access to everything if they pass authorization, they don't have to be in users.conf
    if resource is not None:
        # FIXME: /cobbler/web should not be subject to user check in any case
        for x in ["get", "read", "/cobbler/web"]:
            if resource.startswith(x):
                return 1        # read operation is always ok.

    user_groups = __parse_config()

    # classify the type of operation
    modify_operation = False
    for criteria in ["save", "copy", "rename", "remove", "modify", "edit", "xapi", "background"]:
        if resource.find(criteria) != -1:
            modify_operation = True

    # FIXME: is everyone allowed to copy?  I think so.
    # FIXME: deal with the problem of deleted parents and promotion

    found_user = False
    found_groups = []
    grouplist = list(user_groups.keys())
    for g in grouplist:
        for x in user_groups[g]:
            if x == user:
                found_groups.append(g)
                found_user = True
                # if user is in the admin group, always authorize
                # regardless of the ownership of the object.
                if g == "admins" or g == "admin":
                    return True

    if not found_user:
        # if the user isn't anywhere in the file, reject regardless
        # they can still use read-only XMLRPC
        return 0
    if not modify_operation:
        # sufficient to allow access for non save/remove ops to all
        # users for now, may want to refine later.
        return True

    # Now we have a modify_operation op, so we must check ownership of the object. Remove ops pass in arg1 as a string
    # name, saves pass in actual objects, so we must treat them differently. Automatic installaton files are even more
    # special so we call those out to another function, rather than going through the rest of the code here.

    if resource.find("write_autoinstall_template") != -1:
        return __authorize_autoinst(api_handle, found_groups, user, arg1)
    elif resource.find("read_autoinstall_template") != -1:
        return True

    # The API for editing snippets also needs to do something similar. As with automatic installation files, though
    # since they are more widely used it's more restrictive.

    if resource.find("write_autoinstall_snippet") != -1:
        return __authorize_snippet(api_handle, found_groups, user, arg1)
    elif resource.find("read_autoinstall_snipppet") != -1:
        return True

    obj = None
    if resource.find("remove") != -1:
        if resource == "remove_distro":
            obj = api_handle.find_distro(arg1)
        elif resource == "remove_profile":
            obj = api_handle.find_profile(arg1)
        elif resource == "remove_system":
            obj = api_handle.find_system(arg1)
        elif resource == "remove_repo":
            obj = api_handle.find_repo(arg1)
        elif resource == "remove_image":
            obj = api_handle.find_image(arg1)
    elif resource.find("save") != -1 or resource.find("modify") != -1:
        obj = arg1

    # if the object has no ownership data, allow access regardless
    if obj is None or obj.owners is None or obj.owners == []:
        return True

    return __is_user_allowed(obj, found_groups, user, resource, arg1, arg2)



"""
Authentication module that uses /etc/cobbler/auth.conf
Choice of authentication module is in /etc/cobbler/modules.conf

Copyright 2007-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import hashlib
import os
from typing import List

from cobbler.module_loader import get_module_name


def hashfun(text: str) -> str:
    """
    Converts a str object to a hash which was configured in modules.conf of the Cobbler settings.

    :param text: The text to hash.
    :return: The hash of the text. This should output the same hash when entered the same text.
    """
    hashfunction = get_module_name("authentication", "hash_algorithm", "sha3_512")
    if hashfunction == "sha3_224":
        hashalgorithm = hashlib.sha3_224(text.encode('utf-8'))
    elif hashfunction == "sha3_384":
        hashalgorithm = hashlib.sha3_384(text.encode('utf-8'))
    elif hashfunction == "sha3_256":
        hashalgorithm = hashlib.sha3_256(text.encode('utf-8'))
    elif hashfunction == "sha3_512":
        hashalgorithm = hashlib.sha3_512(text.encode('utf-8'))
    elif hashfunction == "blake2b":
        hashalgorithm = hashlib.blake2b(text.encode('utf-8'))
    elif hashfunction == "blake2s":
        hashalgorithm = hashlib.blake2s(text.encode('utf-8'))
    elif hashfunction == "shake_128":
        hashalgorithm = hashlib.shake_128(text.encode('utf-8'))
    elif hashfunction == "shake_256":
        hashalgorithm = hashlib.shake_256(text.encode('utf-8'))
    else:
        errortext = "The hashfunction (Currently: %s) must be one of the defined in /etc/cobbler/modules.conf!" \
                    % hashfunction
        raise ValueError(errortext)
    return hashalgorithm.hexdigest()


def register() -> str:
    """
    The mandatory Cobbler module registration hook.
    """
    return "authn"


def __parse_storage() -> List[List[str]]:
    """
    Parse the users.digest file and return all users.

    :return: A list of all users. A user is a sublist which has three elements: username, realm and passwordhash.
    """
    if not os.path.exists("/etc/cobbler/users.digest"):
        return []
    with open("/etc/cobbler/users.digest", encoding='utf-8') as fd:
        data = fd.read()
    results = []
    lines = data.split("\n")
    for line in lines:
        try:
            line = line.strip()
            tokens = line.split(":")
            results.append([tokens[0], tokens[1], tokens[2]])
        except:
            pass
    return results


def authenticate(api_handle, username: str, password: str) -> bool:
    """
    Validate a username/password combo.

    Thanks to http://trac.edgewall.org/ticket/845 for supplying the algorithm info.

    :param api_handle: Unused in this implementation.
    :param username: The username to log in with. Must be contained in /etc/cobbler/users.digest
    :param password: The password to log in with. Must be contained hashed in /etc/cobbler/users.digest
    :return: A boolean which contains the information if the username/password combination is correct.
    """

    userlist = __parse_storage()
    for (user, realm, passwordhash) in userlist:
        if user == username and realm == "Cobbler":
            calculated_passwordhash = hashfun(password)
            if calculated_passwordhash == passwordhash:
                return True
    return False



import time


def register() -> str:
    """
    This pure python trigger acts as if it were a legacy shell-trigger, but is much faster. The return of this method
    indicates the trigger type.

    :return: Always `/var/lib/cobbler/triggers/install/pre/*`
    """
    return "/var/lib/cobbler/triggers/install/pre/*"


def run(api, args: list) -> int:
    """
    The method runs the trigger, meaning this logs that an installation has started.

    The list of args should have three elements:
        - 0: system or profile
        - 1: the name of the system or profile
        - 2: the ip or a "?"

    :param api: This parameter is currently unused.
    :param args: Already described above.
    :return: A "0" on success.
    """
    objtype = args[0]
    name = args[1]
    ip = args[2]

    # FIXME: use the logger

    fd = open("/var/log/cobbler/install.log", "a+")
    fd.write("%s\t%s\t%s\tstart\t%s\n" % (objtype, name, ip, time.time()))
    fd.close()

    return 0



"""
Authorization module that allow users listed in
/etc/cobbler/users.conf to be permitted to access resources.
For instance, when using authz_ldap, you want to use authn_configfile,
not authz_allowall, which will most likely NOT do what you want.

This software may be freely redistributed under the terms of the GNU
general public license.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA.
"""


from configparser import SafeConfigParser

import os
from typing import Dict

CONFIG_FILE = '/etc/cobbler/users.conf'


def register() -> str:
    """
    The mandatory Cobbler module registration hook.

    :return: Always "authz".
    """
    return "authz"


def __parse_config() -> Dict[str, dict]:
    """
    Parse the the users.conf file.

    :return: The data of the config file.
    """
    if not os.path.exists(CONFIG_FILE):
        return {}
    config = SafeConfigParser()
    config.read(CONFIG_FILE)
    alldata = {}
    groups = config.sections()
    for g in groups:
        alldata[str(g)] = {}
        opts = config.options(g)
        for o in opts:
            alldata[g][o] = 1
    return alldata


def authorize(api_handle, user: str, resource: str, arg1=None, arg2=None) -> int:
    """
    Validate a user against a resource. All users in the file are permitted by this module.

    :param api_handle: This parameter is not used currently.
    :param user: The user to authorize.
    :param resource: This parameter is not used currently.
    :param arg1: This parameter is not used currently.
    :param arg2: This parameter is not used currently.
    :return: "0" if no authorized, "1" if authorized.
    """
    # FIXME: this must be modified to use the new ACL engine

    data = __parse_config()
    for g in data:
        if user.lower() in data[g]:
            return 1
    return 0



"""
Hard links Cobbler content together to save space.
"""
# SPDX-License-Identifier: GPL-2.0-or-later
# SPDX-FileCopyrightText: Copyright 2009, Red Hat, Inc and Others
# SPDX-FileCopyrightText: Michael DeHaan <michael.dehaan AT gmail>
# SPDX-FileCopyrightText: 2021 Enno Gotthold <egotthold@suse.de>
# SPDX-FileCopyrightText: Copyright SUSE LLC

import logging
import os

from cobbler import utils


class HardLinker:

    def __init__(self, api=None):
        """
        Constructor

        :param api: The API to resolve information with.
        """
        if api is None:
            raise ValueError("cobbler hardlink requires the Cobbler-API for resolving the root folders.")
        self.api = api
        self.hardlink = ""
        self.logger = logging.getLogger()
        self.family = utils.get_family()
        self.webdir = self.api.settings().webdir

        # Getting the path to hardlink
        for possible_location in ["/usr/bin/hardlink", "/usr/sbin/hardlink"]:
            if os.path.exists(possible_location):
                self.hardlink = possible_location
        if not self.hardlink:
            utils.die("please install 'hardlink' to use this feature")

        # Setting the args for hardlink according to the distribution. Must end with a space!
        if self.family == "debian":
            hardlink_args = "-f -p -o -t -v "
        elif self.family == "suse":
            hardlink_args = "-f -v "
        else:
            hardlink_args = "-c -v "
        self.hardlink_cmd = "%s %s %s/distro_mirror %s/repo_mirror" \
                            % (self.hardlink, hardlink_args, self.webdir, self.webdir)

    def run(self):
        """
        Simply hardlinks directories that are Cobbler managed.
        """
        self.logger.info("now hardlinking to save space, this may take some time.")

        utils.subprocess_call(self.hardlink_cmd, shell=True)
        hardlink_cmd = "%s -c -v %s/distro_mirror %s/repo_mirror" % (self.hardlink, self.webdir, self.webdir)

        return utils.subprocess_call(hardlink_cmd, shell=True)
